%
% $Id: heart-alsv.pl,v 1.50 2010-06-24 13:08:01 esimon Exp $
%
% Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
% HeaRT the ALSV(FD) handling
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- write('HeaRT: $Id: heart-alsv.pl,v 1.50 2010-06-24 13:08:01 esimon Exp $'),nl.

%%% Finite domain declaration
%%% domain(<name>,<list-of-values>).

%%% Set value definition
%%% set(<name>,<list-of-values>).

%%% Ordered Set value definition
%%% set(<name>,<list-of-values>,<order-predicate-name>).

%%% Attribute declaration
%%% attribute(<attribute-name>,<attribute-type>,<attribute-domain>).
% attribute(aDD,atomic,day_of_week).
% attribute(sDD,set,day_of_week).


%%% State definition: by listing all the values of the attributes
%%% state(<state-identifier>,<attribute>,<value>,<type>).
%%% state_time(<state-identifier>,<time>).
%%% <type> := atomic|set|...\
% state(s1,aMO,atomic,march).
% state(s2,sTM,set,[7,8,9,10,11,12,13,14,15]).


%%% Definition of facts: to be used in rule preconditions
%%% f(<attribute-type>,<attribute-name>,<relation>,<attribute-domain>)
%%% <attribute-type> := atomic|set|....
%%% For <attribute-type>=atomic <relation> := \eq, \neq, \in, \not\in
%%% For <attribute-type>=set <relation> := \eq, \neq, \subseteq,\supseteq,\sim, \not\sim

%%% To be defined: 'NULL','_',[]


%%% Satisfaction of a single fact versus state: a fact 
%%% f(<attribute-type>,<attribute-name>,<relation>,<attribute-domain>)
%%% is satisfied within some state <state> iff <relation> is valid within this state;
%%% Deatils are defined by the first rows of table1 and table2.
%%% valid(<fact>,<state>).

%%% Table1: cases of \eq, \neq, \in, \notin versus atomic StateValue in State
%valid(f(atomic,A,eq,Value),State) :-
%	state(State,A,atomic,StateValue),
%	Value == StateValue,
%	!.  
%valid(f(atomic,A,neq,Value),State) :-
%	state(State,A,atomic,StateValue),
%	Value =\= StateValue,
%	!. 
%valid(f(atomic,A,in,SetValue),State) :-
%	state(State,A,atomic,StateValue),
%	member(StateValue,SetValue),
%	!. 
%valid(f(atomic,A,notin,SetValue),State) :-
%	state(State,A,atomic,StateValue),
%	\+member(StateValue,SetValue),
%	!. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% heart

alsv_valid(Att eq any, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]),
	heart_debug(2,['Valid:',Att,eq,any]).

alsv_valid(Att in [any], State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]),
	heart_debug(2,['Valid:',Att,in,[any]]).
	
alsv_valid(Att eq any, State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	heart_debug(2,['Valid:',Att,eq,any]).

alsv_valid(Att eq [any], State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	heart_debug(2,['Valid:',Att,eq,[any]]).

alsv_valid(Att eq null, State) :-
	%alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	\+ alsv_values_get(State,Att,StateValue), %No values for this state
	heart_debug(2,['Valid:',Att,eq,null]).
	
alsv_valid(Att eq null, State) :-
	%alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	StateValue = null, 					   %State value explicite set to null.
	heart_debug(2,['Valid:',Att,eq,null]).

alsv_valid(Att in [L to U], State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[L]), alsv_values_check(Att,[U]),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,[L],[NormL]),
	normalize(Att,[U],[NormU]),
	NormStateValue =< NormU,
	NormStateValue >= NormL,		
	heart_debug(2,['Valid:',Att,in,[L to U]]).
	
alsv_valid(Att notin [L to U], State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[L]), alsv_values_check(Att,[U]),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,[L],[NormL]),
	normalize(Att,[U],[NormU]),
	NormStateValue >= NormU,
	NormStateValue =< NormL,	
	heart_debug(2,['Valid:',Att,notin,[L to U]]).
	
alsv_valid(Att in SetValue, State) :-		
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), 	
	alsv_values_check(Att,SetValue),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,SetValue,NormSetValue),
	(member(NormStateValue,NormSetValue);math_member(NormStateValue,NormSetValue)),		%or math_member
	heart_debug(2,['Valid:',Att,in,SetValue]).
	
alsv_valid(Att notin SetValue, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,SetValue),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,SetValue,NormSetValue),
	\+ member(NormStateValue,NormSetValue),
	\+ math_member(NormStateValue,NormSetValue),
	heart_debug(2,['Valid:',Att,notin,SetValue]).
	
	
alsv_valid(Att eq Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue == NormStateValue,
	heart_debug(2,['Valid:',Att,eq,Value]).
	
alsv_valid(Att neq Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue == NormStateValue,
	heart_debug(2,['Valid:',Att,neq,Value]).

alsv_valid(Att gt Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),	
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue < NormStateValue,
	heart_debug(2,['Valid:',Att,gt,Value]).
	
alsv_valid(Att gte Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),	
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue =< NormStateValue,
	heart_debug(2,['Valid:',Att,gte,Value]).
	
alsv_valid(Att lt Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue > NormStateValue,
	heart_debug(2,['Valid:',Att,lt,Value]).
	
alsv_valid(Att lte Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue >= NormStateValue,
	heart_debug(2,['Valid:',Att,lte,Value]).
	
%%% Table2: cases of \eq, \neq, \subseteq, \supseteq,\sim, \notsim  versus set StateValue in State
alsv_valid(Att eq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	eqset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,eq,Set]).
	
alsv_valid(Att neq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	neqset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,eq,Set]).
	
alsv_valid(Att subseteq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	alsv_subset(NormStateValue,NormSet),
	heart_debug(2,['Valid:',Att,subset,Set]).
	
alsv_valid(Att supseteq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	alsv_subset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,supset,Set]).
	
alsv_valid(Att sim Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	simset(NormStateValue,NormSet),		%myIntersection!
	heart_debug(2,['Valid:',Att,sim,Set]).
	
alsv_valid(Att notsim Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	\+simset(NormStateValue,NormSet),		%myIntersection!
	heart_debug(2,['Valid:',Att,notsim,Set]).

alsv_valid(F,_) :-
	heart_debug(2,['Check for:',F,'NYI!']),
	fail.
	
alsv_values_get(State,Att,StateValue) :-
	xstat State: [Att,StateValue],!.
	
alsv_values_get(_,Att,StateValue) :-
	xattr(A),
	member(name:Att,A),
	member(comm:C,A),
	(C=in;C=inter),
	fire_callback(Att),			% Callback firing - only when infrerence was called as goxio?
	xstat current: [Att,StateValue].

%valid(f(set,A,\subseteq,SetValue),State) :-
%	state(State,A,set,StateValue),
%	subset(SetValue,StateValue),
%	!.
%valid(f(set,A,\supseteq,SetValue),State) :-
%	state(State,A,set,StateValue),
%	subset(StateValue,SetValue),
%	!.
%valid(f(set,A,\sim,SetValue),State) :-
%	state(State,A,set,StateValue),
%	intersect(SetValue,StateValue,[_|_]),
%	!.
%valid(f(set,A,\notsim,SetValue),State):-
%	state(State,A,set,StateValue),
%	intersect(SetValue,StateValue,[]),
%	!.
	


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% alsvd checks
alsv_values_check(_,[]).

%set state value check
% alsv_values_check(Att,[LS to US|Rest]) :-
	% alsv_domain(Att,[LD to UD],numeric), 
	% heart_debug(2,['Check:',Att,numericreg,LS to US]),		
	% LS >= LD,
	% US =< UD,	
	% alsv_values_check(Att,Rest),
	% !.
	
%domain is set, but value X to Y
alsv_values_check(Att,[LS to US|Rest]) :-
	alsv_domain(Att,Domain,numeric), 
	heart_debug(2,['Check:',Att,numericreg,LS to US]),		
	%normalize(Att,Domain,NDomain),	
	%sort(NDomain, [First|SNDomain]),
	%last(SNDomain,Last),
	%LS >= First,
	%US =< Last,
	math_member([LS to US],Domain),
	alsv_values_check(Att,Rest),
	!.
	
%alsv_values_check(Att,[StateValue|Rest]) :-
%	 nonvar(StateValue),
%	 alsv_domain(Att,[LD to UD],numeric),
%	 heart_debug(2,['Check:',Att,numericreg,StateValue]),
%	 number(StateValue),
%	 StateValue >= LD,
%	 StateValue =< UD,	
%	 alsv_values_check(Att,Rest),
%	 !.

	
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),	
	alsv_domain(Att,Domain,numeric),
	heart_debug(2,['Check:',Att,numeric,StateValue]),	
	math_member(StateValue,Domain), % check if value in the domain with math_memeber
	alsv_values_check(Att,Rest),
	!.
	
%domain is set, but value X to Y
alsv_values_check(Att,[LS to US|Rest]) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,symbolic), 
	heart_debug(2,['Check:',Att,symbolic,LS to US]),		
	normalize(Att,Domain,NDomain),	
	normalize(Att,LS,NLS),
	normalize(Att,US,NUS),
	sort(NDomain, [First|SNDomain]),
	last(SNDomain,Last),
	NLS >= First,
	NUS =< Last,
	alsv_values_check(Att,Rest),
	!.
		
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	number(StateValue),
	math_member(StateValue,Domain), %chceck if weight in the domain with math_memeber
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	member(StateValue/_,Domain),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	number(StateValue),
	truncate(StateValue) =:= StateValue,
	TruncatedValue is truncate(StateValue),
	nth1(TruncatedValue,Domain,_),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
%	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	member(StateValue,Domain),
	alsv_values_check(Att,Rest),
	!.

% alsv_values_check for general atts missing! ex. [1,2,3,4,5] domain [1,2,3,4,5,6,7,8]. subset

alsv_values_check(Att,[V|_]) :-
	write('  ERROR: values of \"'), write(Att),
	write('\": '), write(V), write(' NOT in the domain!'),nl,!.

%retrieves attribute domain, and base: symbolic, numeric
alsv_domain(Att,Domain,Base) :-
	xattr(A),
	member(name:Att,A),
	member(type:T,A),
	xtype(Ts),
	member(name:T,Ts),
	member(domain:Domain,Ts),
%	alsv_attr_class(Att,simple),
	member(base:Base,Ts).

%alsv_domain(Att,ExpandedDomain,Base) :-
%	xattr(A),
%	member(name:Att,A),
%	member(type:T,A),
%	xtype(Ts),
%	member(name:T,Ts),
%	member(domain:Domain,Ts),
%	alsv_attr_class(Att,general),
%	expand_range_expression(Domain,ExpandedDomain), %for general attributesthat makes sense
%	member(base:Base,Ts).

% retrieves attribute class: simple, general
alsv_attr_class(Att,Class) :-
	xattr(A),
	member(name:Att,A),
	member(class:Class,A).

alsv_order(Att,yes) :-
	xattr(A),
	member(name:Att,A),
	member(type:T	,A),
	xtype(Ts),	
	member(name:T,Ts),
	member(base:numeric, Ts),!.

alsv_order(Att,Order) :-
	xattr(A),
	member(name:Att,A),
	member(type:T,A),
	xtype(Ts),
	member(name:T,Ts),
	member(ordered:Order,Ts),!.

alsv_order(_,no).
	
normalize(Att, Value, Output) :-
	\+ alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	member(Value,Domain),
	Value = Output,!.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	number(Value),			%and change output to floats if number (can be outside the domain)
	alsv_domain(Att,Domain,_),
	math_member(Value,Domain),	
	Output is Value*1.0,!.

normalize(Att, Value, Output) :-	
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	member(Value/Numbering,Domain), 
	Output is Numbering*1.0,!.

normalize(Att, MixedValue, Output) :-	
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	member(MixedValue,Domain),
	MixedValue = _/Numbering, 
	Output is Numbering*1.0,!.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	atom(Value),			%and change output to floats if number
	nth1(Number,Domain,Value),
	Output is Number*1.0.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	number(Value),
	truncate(Value) =:= Value,
	TruncatedNumber is truncate(Value),
	nth1(TruncatedNumber,Domain,_),	%and change output to floats
	Output is TruncatedNumber*1.0,!.

normalize(Att,[X to Y | Rest1],Rest2) :-
	alsv_domain(Att,_,numeric),	
	normalize(Att,X,XN),
	normalize(Att,Y,YN),
	normalize(Att,Rest1,Rest),
	merge([XN to YN],Rest,Rest2),
	!.

% Probably it's not useful
normalize(Att, [X to Y| Rest1], Rest2) :-
	alsv_order(Att,yes),		
	alsv_domain(Att,MixedDomain,symbolic),
	normalize(Att,X,XN),
	normalize(Att,Y,YN),
	CXN is ceiling(XN),
	FYN is floor(YN),
	findall(FV,(between(CXN,FYN,V),FV is V*1.0),InsideList),		
	merge([XN|InsideList],[YN],DuplicatedList),
	list_to_set(DuplicatedList,List),
	expand_range_expression(MixedDomain,Domain),
	normalize(Att,Domain,NDomain),						% Due to expand_range_exp we are sure tht no inf loop occure
	intersection(NDomain, List, FilteredList),
	normalize(Att, Rest1, Rest),
	merge(FilteredList, Rest,Rest2),
	!.

normalize(Att,[Number|Rest1],[Value|Rest2]) :-
	normalize(Att,Number,Value),
	normalize(Att,Rest1,Rest2),
	!.
	
normalize(_,[],[]).

%only for mixed domains ex [12,3,7 to 10, 44].
expand_range_expression([],[]).
expand_range_expression([X to Y|Values],Expanded) :-
	Start is ceiling(X),
	Stop is floor(Y),
	findall(V,(between(Start,Stop,V)),InsideList),	
	merge([X|InsideList],[Y],DuplicatedList),
	list_to_set(DuplicatedList,List),
	expand_range_expression(Values,Rest),
	merge(List,Rest,Expanded),!.
expand_range_expression([H|Values],[H|Expanded]) :-
	nonvar(H),
	expand_range_expression(Values,Expanded),!.
	

	
	
%normalize(Att, Value, Output) :-
%	alsv_order(Att,yes),
%	findall(Numbering,member(Name/Numbering,Value),Output), %and change output to floats
%	Output \= [],
%	!.	

math_member(Number, [Lower to Upper|_]) :-
	number(Number),
	number(Lower),
	number(Upper),
	Number =< Upper,
	Number >= Lower,!.


math_member(Number, [Lower to Upper|Rest]) :-
	number(Number),
	number(Lower),
	number(Upper),
	math_member(Number,Rest),!.

math_member(L to U, Set) :-
	is_list(Set),
	member(L,Set),
	member(U,Set),!.

math_member(L to U,[Number|Rest]) :-
	number(Number),
	math_member(L to U,Rest),!.

math_member(L to U,[SL to SU|_]) :-
	number(L),
	number(U),
	number(SL),
	number(SU),
	L >= SL,
	U =< SU,!.

math_member(L to U,[SL to SU|Rest]) :-
	number(L),
	number(U),
	number(SL),
	number(SU),
	math_member(L to U,Rest),!.


math_member(Weight,[_/DomainWeight|[]]) :-
	number(Weight),
	number(DomainWeight),
	Weight =:= DomainWeight,!. 

math_member(Weight,[_/DomainWeight|Rest]) :-
	number(Weight),
	number(DomainWeight),
	Weight =\= DomainWeight, 
	math_member(Weight,Rest),!.
	
math_member(Weight,[_/DomainWeight|_]) :-
	number(Weight),
	number(DomainWeight),
	Weight =:= DomainWeight,!.
	
math_member(Number,[Head|[]]) :-
	number(Number),
	number(Head),
	Number =:= Head,!. 
math_member(Number,[Head|Rest]) :-
	number(Number),
	number(Head),
	Number =\= Head, 
	math_member(Number,Rest),!.
	
math_member(Number,[Head|_]) :-
	number(Number),
	number(Head),
	Number =:= Head,!.


math_member([],_).

math_member([Head|Rest],Set) :-
	math_member(Head,Set),
	math_member(Rest,Set),!.

	


%%% Auxiliary predictaes
%%% Set equality definition:
%%% !!! We assume that both sets are normalized !!!
alsv_subset(Set1, Set2) :-
	subset(Set1,Set2),!.
alsv_subset(Set1,Set2) :-
	math_member(Set1,Set2),!.

eqset(Set1,Set2) :-
	alsv_subset(Set1,Set2),
	alsv_subset(Set2,Set1),!.
%%% Set in-equality definition:
neqset(Set1,Set2) :-
	\+ eqset(Set1,Set2).

simset(Set1, Set2) :-
	\+ sub(Set1, Set2, Set1).


onesub([],_,[]).	

%for numbers

onesub([L1 to U1|Set1], [L2 to U2],[L1 to NU1, NL2 to U1|R]) :-
	L2 > L1,
	U2 < U1,
	NU1 is L2-1,
	NL2 is U2+1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1], [L2 to U2],[NL1 to U1|R]) :-
	U2 =:= L1,
	U1 > U2,
	NL1 is L1+1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1], [L2 to U2],[L1 to NU1|R]) :-
	U1 =:= L2,
	U2 > U1,
	NU1 is U1-1,
	onesub(Set1,[L2 to U2],R),!.


onesub([L1 to U1|Set1], [L2 to U2],[L1 to NU1|R]) :-
        L1 < L2,
	L2 < U1,
	U2 >= U1,
	NU1 is L2-1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1], [L2 to U2],[NL1 to U1|R]) :-
	U1 > U2,
	L1 < U2,
	L2 =< L1,
	NL1 is U2+1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1],[Number],[NL to U1|R]) :-
	L1 \= U1,
	number(Number),
	Number =:= L1,
	NL is Number+1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[L1 to NU|R]) :-
	L1 \= U1,
	number(Number),
	Number =:= U1,
	NU is Number-1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[L1 to NU, NL to U1|R]) :-
	L1 \= U1,
	number(Number),
	math_member(Number, [L1 to U1]),
	NU is Number-1,
	NL is Number+1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[L1 to U1|R]) :-
	L1 \= U1,
	number(Number),
	\+math_member(Number, [L1 to U1]),
	onesub(Set1, [Number],R),!.

onesub([Number|Set1],[L2 to U2],[Number|R]) :-
	number(Number),
	\+ math_member(Number, [L2 to U2]),
	onesub(Set1, [L2 to U2],R),!.

onesub([Number1|Set1],[Number2],[Number1|R]) :-
	number(Number1),
	number(Number2),
	Number1 =\= Number2,	
	onesub(Set1, [Number2],R),!.

%for symbolic and disjoint sets X to Y!!!! 
onesub([Element1|Set1],[Element2],[Element1|R]) :-
	\+number(Element1),
	\+number(Element2),
	Element1 \= Element2,
	onesub(Set1,[Element2],R),!.



onesub( [_| Set1],E,R) :-
	onesub(Set1,E,R),!.

sub(R,[],R).

sub(Set1,[H|Set2],R) :-
	onesub(Set1, [H], RR),
	sub(RR,Set2, R),!.
	
alsv_union(Set1, Set2, Res) :-
 	merge(Set1, Set2,DRes),
	list_to_set(DRes, Res),!.

alsv_intersect(Set1, Set2, Res) :-
	sub(Set1,Set2,R),
	sub(Set1,R,Res),!.

alsv_except(Set1, Set2, Res) :-
	sub(Set1, Set2, Res),!.

alsv_completment(Att, Set1, Res) :-
	alsv_domain(Att, Domain, _),
	normalize(Att, Domain, NDomain),
	alsv_attr_class(Att, general),
	sub(NDomain, Set1, Res),!.

alsv_eval_general(Expr, Expr) :-
	is_list(Expr),!.

alsv_eval_general(Expr, Eval) :-
	Expr = union(V1,V2),
	alsv_eval_general(V1,R1),
	alsv_eval_general(V2,R2),
	alsv_union(R1,R2,Eval),!.

alsv_eval_general(Expr, Eval) :-
	Expr = intersect(V1,V2),
	alsv_eval_general(V1,R1),
	alsv_eval_general(V2,R2),
	alsv_intersect(R1,R2,Eval),!.

alsv_eval_general(Expr, Eval) :-
	Expr = except(V1,V2),
	alsv_eval_general(V1,R1),
	alsv_eval_general(V2,R2),
	alsv_except(R1,R2,Eval),!.

alsv_eval_general(Expr, Eval) :-
	Expr = complement(Att,V2),
	alsv_eval_general(V2,R2),
	alsv_complement(Att,R2,Eval),!.

	
	

% heart
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

alsv_transition(Att set Expr) :-
%	atom(Val),
	\+ Expr = [_|_],
	alsv_domain(Att,_,numeric),
	alsv_attr_class(Att,simple),
	alsv_eval_simple(Expr, Val),
	heart_debug(2,['Adding new fact: ',Att,' set to ',Val]),
	alsv_new_val(Att,Val).

alsv_transition(Att set Val) :-
%	atom(Val),
	\+ Val = [_|_],
	alsv_domain(Att,_,symbolic),
	alsv_attr_class(Att,simple),
	heart_debug(2,['Adding new fact: ',Att,' set to ',Val]),
	alsv_new_val(Att,Val).

alsv_transition(Att set Expr) :-
%	atom(Val),
	Expr = [_|_],
	alsv_eval_general(Expr,Val),
	alsv_attr_class(Att,general),
%	alsv_values_check(Att,[Val]),
	alsv_new_val(Att,Val).

% fixme: no direct assert/retract, but set/get_att_val
alsv_new_val(A,V):-
%	xstat current:[A,O],
%	alsv_values_check(A,[V]),
	retractall(xstat current:[A,_]),
	assert(xstat current:[A,V]),
	heart_debug(2,['Trans:',A,to,V]).
%	heart_debug(['Transition: ',A,from,O,to,V]).

alsv_eval_simple(Expr, Evaluated) :-
	parse(Expr, Parsed),
	Resolved =.. Parsed,
	Evaluated is Resolved.

parse(Expr, Parsed) :-
	Expr =.. List,
	parse_helper(List,Parsed),!.	

parse_helper([],[]).

parse_helper([H|Rest],[Val|Parsed]) :-
	parse_is_attr(H),
	alsv_values_get(current,H,Val),
	parse_helper(Rest,Parsed),!.
	
parse_helper([H|Rest],[H|Parsed]) :-
	number(H),
	parse_helper(Rest, Parsed),!.	

parse_helper([H|Rest],[H|Parsed]) :-
	parse_is_operator(H),
	parse_helper(Rest,Parsed),!.

parse_helper([SubExpr|Rest], [Unsplit|Parsed]) :-
	SubExpr =.. Split,
	parse_helper(Split,Subparsed),
	Unsplit =.. Subparsed,
	parse_helper(Rest, Parsed),!.


parse_is_operator(H) :-
	H = '-';
	H = '+';
	H = '*';
	H = '/';
	H = ')';
	H = ')'.

parse_is_attr(H) :-
	xattr(P),
	member(name:H,P),!.
	
	

%alsv_transition(Att set Val) :-
%	Val = [_|_],
	
%  [therm_set set 24].
%  [therm_set set [24,5]].

%  [therm_set set seval union([24,5],intersect('att_ala',[4 to 8])). EXCEPT COMPLEMENT? % binary relations! result=set

%  [therm_set set aeval 4 + 7 * 'att_ala']. %dowolone wyr oblicznae przez is, maths builtin, defined funcs? % result=number binary relations! therm_set in numeric

  
%%% Checking if a list of facts -- precondition formula is satisfied 
%%% satisfied(<list-of-facts>,State).

%satisfied([],_) :- !.
%satisfied([Fact|List],State) :-
%	valid(Fact,State),
%	satisfied(List,State).

%%% Auxiliary predictaes
%%% Set equality definition:
%eqset(Set1,Set2) :-
%	subset(Set1,Set2),
%	subset(Set2,Set1).
%%% Set in-equality definition:
%neqset(Set1,Set2) :-
%	\+ eqset(Set1,Set2).




