%FIXME - removing spaces
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DAAL prototype translator.
% --------------------------
%
% Translates attributes into concepts, 
%  attribute values into instances and
%  rule preconditions into T-Box like axioms.
%   
% Supported operators: 'in', 'eq'.
% Supported attributes: 'symbolic'.
% 
% Basic predicates:
%   owl_xml_gen/0 - translate HMR file into DAAL representation (ontology).
%   owl_xml_gen/1 - translate HMR file into DAAL representation (ontology)
%                   and write it to the file given as the argument.
%   owl_xml_attr_gen/0 - translate the attribute definitions
%   owl_rulp_gen/0 - translate the rules preconditions
%   owl_rulp_gen/1 - translate the given rule preconditions
%   owl_stat_gen/0 - translate the states statements
%   owl_stat_gen/1 - translate the given state statement
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%:- ensure_loaded('heart.pl').
%:- ensure_loaded('therm-rt.pl').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- use_module(library(lists)).

owl_xml_gen(F) :-
	tell(F),
	owl_xml_gen,
	told.

owl_xml_gen :-
	xml_indent_mark_set('owl'),
	xml_indent_clear('owl'),
	owl_xml_head,

        owl_xml_attr_gen,
        owl_xml_rulp_gen,
        owl_xml_stat_gen,
        owl_xml_allDifferent,

	owl_xml_tail,
	!.

owl_xml_head :-
    write('<?xml version="1.0"?>'), nl,
    write('<rdf:RDF'), nl,
    write('    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"'), nl,
    write('    xmlns:protege="http://protege.stanford.edu/plugins/owl/protege#"'), nl,
    write('    xmlns:xsp="http://www.owl-ontologies.com/2005/08/07/xsp.owl#"'), nl,
    write('    xmlns:owl="http://www.w3.org/2002/07/owl#"'), nl,
    write('    xmlns:xsd="http://www.w3.org/2001/XMLSchema#"'), nl,
    write('    xmlns:swrl="http://www.w3.org/2003/11/swrl#"'), nl,
    write('    xmlns="http://www.owl-ontologies.com/Ontology1280240317.owl#"'), nl,
    write('    xmlns:swrlb="http://www.w3.org/2003/11/swrlb#"'), nl,
    write('    xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"'), nl,
    write('  xml:base="http://www.owl-ontologies.com/Ontology1280240317.owl">'), nl,
    write('  <owl:Ontology rdf:about=""/>'), nl,

	xml_indent_more('owl'),
	write('<owl:Annotation>'), nl,
	xml_indent_more('owl'),
	write('<owl:Documentation> HMR to OWL "DAAL" Translator</owl:Documentation>'), nl,
	xml_indent_less('owl'),
	write('</owl:Annotation>'), nl,
%	write('<owl:VersionInfo>v 1.0 2010/08/05 Weronika T. Adrian wta@agh.edu.pl</owl:VersionInfo>'), nl,
	nl.
	
owl_xml_tail :-
    write('</rdf:RDF>'),
    nl, nl,
    xml_indent_less('owl').


%%%%%%%%%%%%%%%%%%%%
% Attributes
%%%%%%%%%%%%%%%%%%%%
owl_xml_attr_gen :-
    owl_xml_attr_gen(_).

owl_xml_attr_gen(A):-
    xattr(A),
    member(name:Name,A),
    member(type:TypeName,A),
    xtype(Type),
    member(name:TypeName,Type),
    member(base:symbolic,Type),
    member(domain:Domain,Type),
    owl_xml_attr_gen2(Name,Domain),
    fail.

owl_xml_attr_gen(A):-
    xattr(A),
    member(name:Name,A),
    member(type:TypeName,A),
    xtype(Type),
    member(name:TypeName,Type),
    member(base:numeric,Type),
    write('<!-- ***WARNING: Attribute '),
    write(Name),
    write(' is of a numeric type: it is not handled in this translation.*** -->' ),nl,nl.

owl_xml_attr_gen(_).


owl_xml_attr_gen2(Name,Domain) :-
    write('<!-- Definition of the attribute: '),
    write(Name),
    write(' -->'),nl,
	xml_tag_open_indent('owl:Class',[['rdf:ID',Name]]),nl,
	xml_tag_open_indent('rdfs:subClassOf',_),nl,
	xml_tag_open_indent('owl:Class',_),nl,
	xml_tag_open_indent('owl:oneOf',[['rdf:parseType','Collection']]),nl,
        xml_list_values(Domain),
	xml_tag_close_indent('owl:oneOf'),
	xml_tag_close_indent('owl:Class'),
	xml_tag_close_indent('rdfs:subClassOf'),
	xml_tag_close_indent('owl:Class'),nl.

xml_list_values([Value|Rest]):-
    write('     <rdf:Description rdf:about="#'),
    write(Value),
    write('" />'),nl, 
    xml_list_values(Rest).

xml_list_values([]).


xml_list_values(Value):-
%    \+compound(Value),
    atom(Value),Value\=[],
    concat_atom(['#',Value],Val),
    xml_tag_single_indent('rdf:Description',[['rdf:about',Val]]).

%xml_list_values([Value|Rest]):-
%    atomic(Value),
%    concat_atom(['#',Value],Val),
%    xml_tag_single_indent('rdf:Description',[['rdf:about',Val]]),
%    xml_list_values(Rest).
%xml_list_values([]).
%xml_list_values(Value):-
%%    \+compound(Value),
%    atom(Value),Value\=[],
%    concat_atom(['#',Value],Val),
%    write('Pojedyncza wartość'),
%    xml_tag_single_indent('rdf:Description',[['rdf:about',Val]]).
%xml_list_values([Value|Rest]):-
%    compound(Value),
%%    Value= to(V,V2),
%    write('***WARNING: ranges in rule conditions e.g. attr in [9 to 17] not supported.'),nl.
%    xml_list_values(Rest).


%%%%%%%%%%%%%%%%%%%%
% Rule preconditions
%%%%%%%%%%%%%%%%%%%%

owl_xml_rulp_gen:-
    owl_xml_rulp_gen(_).

owl_xml_rulp_gen(R):-
    xrule(R:Cond==>_),
    write('<!-- Preconditions of the rule:'),
    write(R),write(' -->'),nl,
    owl_xml_rule_cond_gen(Cond),fail.
owl_xml_rulp_gen(_).

    
owl_xml_rule_cond_gen([]).
owl_xml_rule_cond_gen([C|Rest]) :-
    owl_xml_daal_axiom(C),
    owl_xml_rule_cond_gen(Rest).

%Supported operators: 'in', 'eq'
owl_xml_daal_axiom(C):-
    arg(1,C,Attribute),
    functor(C,Operator,_),
    member(Operator,['in','eq']),
    arg(2,C,any),
    write('<!-- The precondition in HMR:'),
    write(C),write(' - so no DAAL axiom is generated.-->'),nl,nl,!.

owl_xml_daal_axiom(C):-
    arg(1,C,Attribute),
    functor(C,Operator,_),
    member(Operator,['in','eq']),
    arg(2,C,[to(From,To)]),
    write('<!-- The precondition in HMR:'),
    write(C),write(' -->'),nl,
    write('<!-- **WARNING: ranges are not supported! -->'),nl,nl,!.

owl_xml_daal_axiom(C):-
    arg(1,C,Attribute),
    functor(C,Operator,_),
    member(Operator,['in','eq']),
    arg(2,C,Values),
    write('<!-- The precondition in HMR:'),
    write(C),write(' -->'),nl,
    xml_tag_open_indent('owl:Class',[['rdf:ID',Attribute]]),nl,
    xml_tag_open_indent('owl:equivalentClass',_),nl,
    xml_tag_open_indent('owl:Class',_),nl,
    xml_tag_open_indent('owl:oneOf',[['rdf:parseType','Collection']]),nl,
    xml_list_values(Values),
    xml_tag_close_indent('owl:oneOf'),
    xml_tag_close_indent('owl:Class'),
    xml_tag_close_indent('owl:equivalentClass'),
    xml_tag_close_noindent('owl:Class'),nl.


%%%%%%%%%%%%%%%%%%%%
% States
%%%%%%%%%%%%%%%%%%%%

owl_xml_stat_gen:-
    owl_xml_stat_gen(_).
    
owl_xml_stat_gen(State):-
    xstat State:[Attr,Value],
    write('<!-- Definition of the state: '),
    write(State),write(': ['),write(Attr),
    write(','),write(Value),write('] -->'),nl,
    xml_tag_single_indent(Attr,[['rdf:ID',Value]]),
    fail.
owl_xml_stat_gen(_).    



%%%%%%%%%%%%%%%%%%%%
% Additional axiom:
% - make all the possible attributes values different from each other,
% - operator "to" is not supported
%%%%%%%%%%%%%%%%%%%%

owl_xml_allDifferent:-
    %find all attributes and their possible values (domains)
    setof([Attr,Domain],attrDomains(Attr,Domain),List),
    %create a list
    %make head
    write('<owl:AllDifferent>'), nl,
    write('  <owl:distinctMembers rdf:parseType="Collection">'), nl,
    makeAttrValueList(List),
    %for each item make a line
    %make tail.
    write('  </owl:distinctMembers>'), nl,
    write('</owl:AllDifferent>'), nl.
    
attrDomains(Attr,Domain):-
    xattr(A),
    member(name: Attr,A),
    member(type: Type,A),
    xtype(T),
    member(name: Type,T),
    member(domain: Domain,T).

makeAttrValueList([]).
makeAttrValueList([H|T]):-
    makeAttrValueTag(H),
    makeAttrValueList(T).


makeAttrValueTag([A,[to(From,To)|T]]):-
    write('<!-- **WARNING: ranges are not supported! -->'),nl,
    write('<!-- '),write(A),write(': ['),
    write(From),write(' to '),write(To),write('] -->'),nl,
    makeAttrValueTag([A,T]),!.

makeAttrValueTag([A,[V|T]]):-
    write('   <'),
    write(A),
    write(' rdf:about="#'),
    write(V),
    write('" />'),nl, 
%    xml_tag_single_indent(A,[['rdf:about',V]]),
    makeAttrValueTag([A,T]).
makeAttrValueTag([A,[]]).
        

% General XML handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

xml_tag_single(Tag,Atts,Out) :-
	var(Atts),
	concat_atom(['<',Tag,'>'],Out).
xml_tag_single(Tag,Atts,Out) :-
	xml_atts_process(Atts,AttsProc),
	concat_atom(['<',Tag,AttsProc,'/>'],Out).
xml_tag_open(Tag,Atts,Out) :-
	var(Atts),
	concat_atom(['<',Tag,'>'],Out).
xml_tag_open(Tag,Atts,Out) :-
	xml_atts_process(Atts,AttsProc),
	concat_atom(['<',Tag,AttsProc,'>'],Out).
xml_tag_close(Tag,Out) :-
	concat_atom(['</',Tag,'>'],Out).

xml_tag_open_indent(Tag,Atts) :-
	xml_indent_current(CI),
	xml_indent(CI,L),
	xml_tag_open(Tag,Atts,To),!, %%% bug work-around
	concat(L,To,TagOpen),
	write(TagOpen).

xml_tag_single_indent(Tag,Atts) :-
	xml_indent_current(CI),
	xml_indent(CI,L),
	xml_tag_single(Tag,Atts,To),
	concat(L,To,TagOpen),
	write(TagOpen),
	nl.

xml_tag_close_indent(Tag) :-
	xml_indent_current(CI),
	xml_indent(CI,L),
	xml_tag_close(Tag,To),
	concat(L,To,TagOpen),
	write(TagOpen),
	nl.

xml_tag_close_noindent(Tag) :-
	xml_tag_close(Tag,To),
	write(To),
	nl.

xml_content(C) :-
	xml_indent_current(CI),
	xml_indent(CI,L),
	concat(L,C,TagOpen),
	write(TagOpen),
	nl.

% list of pairs, name,value
xml_atts_process([],'').
xml_atts_process([A1|A],P) :-
	A1=[N,V],
	xml_att_build(N,V,P1),
	xml_atts_process(A,P2),
	concat(P1,P2,P).
xml_att_build(N,V,P) :-
	concat_atom([' ',N,'=','"',V,'"'],P).

:-dynamic(xml_indent/2).
:-dynamic(xml_indent_current/1).

xml_indent_clear(Mark) :-
	retractall(xml_indent(Mark,_)),
	asserta(xml_indent(Mark,'')).
xml_indent_more(Mark) :-
	xml_indent(Mark,''),
	retractall(xml_indent(Mark,_)),	
	asserta(xml_indent(Mark,'    ')).
xml_indent_more(Mark) :-
	xml_indent(Mark,L),
	concat('    ',L,N),
	retractall(xml_indent(Mark,_)),
	asserta(xml_indent(Mark,N)).
xml_indent_less(Mark) :-
	xml_indent(Mark,'').
xml_indent_less(Mark) :-
	xml_indent(Mark,L),
	concat('    ',N,L),
	retractall(xml_indent(Mark,_)),
	asserta(xml_indent(Mark,N)).
	
xml_indent_mark_set(M) :-
	retractall(xml_indent_current(_)),
	asserta(xml_indent_current(M)).
