%
% $Id: heart-infer.pl,v 1.13 2009-11-10 07:19:17 esimon Exp $
%
%     Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
%     HeaRT inference procedures
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2009   by S. Bobek, M. Gawedzki
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- write('HeaRT: $Id: heart-infer.pl,v 1.13 2009-11-10 07:19:17 esimon Exp $'),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Single Table processing
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% table processing
runtables([]).
runtables([R|T]) :-
	runtable(R),
	trajectory_projection(R,table), 
	runtables(T).

runtable(T) :-
%	write('*Firing tabl: \"'),
%	write(T), write('\" ('),
	%FIXME: conditionless and decisionless rules support!!!
	%xschm(T:X==>Y),
	xtable(T,_,X,Y),
%	write(X),write('==>'),write(Y),write(')'),nl,
	heart_debug(1,['Firing tabl: ','(',T,')',X,'==>',Y]),
	runrules(T),
	nl.

runrules(T):-
	xrule(T/R:C==>D),
%	write(' *Firing rule: '), write(T/R),nl,
	heart_debug(1,['Firing rule: ',T/R]),
	cond_satisfy(C),
	process_decision(D,Go),
	inference_transfer(Go),
	trajectory_projection(T/R,full), 
	fail.
runrules(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GDI-Firing
%inference: GoalDriven GDI
%
%    * get set of goal tables
%    * build a sequential execution plan, by backward chaining using TT links, list of needed components
%    * run independent tables
%    * wave-like in steps
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Goal driven for specified tables. Tables shoul be a list/set ex.: runrules_GDI([os]).
runrules_GDI(Goals) :-
	create_stack(OrderedTables,Goals),
	write(OrderedTables),nl,
	runtables(OrderedTables).
	
% Runs all tables in order they are linked in fact - this implements DDI inference
runrules_GDI :-
	get_goal_tables(Goals),
	write('Goals: '), write(Goals),nl,
	create_stack(OrderedTables, Goals),
	runtables(OrderedTables).

create_stack([],[]).
create_stack(Stack,[HG|RG]) :-
		create_stack_helper(FrontStack, [HG]),
		create_stack(RareStack,RG),
		join_tables(FrontStack,RareStack,StackDuplicated),
		list_to_set(StackDuplicated,Stack),!.
	
create_stack_helper(Stack,Goal) :-
	get_pred_tables(Goal,Predecessors),
	join_tables(Predecessors,Goal,Stack).

get_pred_tables([],[]).
get_pred_tables(Goals,Predecessors) :-
	findall(C,condition_goal(C,Goals),Conds),
	flatten(Conds,FC),
	findall(T,decision_goal(T,FC),Preds),
	flatten(Preds,FP),
	get_pred_tables(FP,FPP),
	join_tables(FPP,FP,Predecessors), !.

join_tables([],X,X).
join_tables([X|L1],L2,[X|L3]) :-
	join_tables(L1,L2,L3).


condition_goal(C,Set) :-
	%xschm(A:C==>_),
	xtable(A,_,C,_),
	member(A,Set).
		
decision_goal(T,Set) :-
	%xschm(T:_==>D),
	xtable(T,_,_,D),
	nonvar(Set),
	intersection(D,Set,[_|_]). %If we assume that there may be links to more than one table.
	
decision_goal(T,Set) :-
	%xschm(T:_==>D),
	xtable(T,_,_,D),
	var(Set),
	Set=D.

find_goal_atts(GoalAtts) :-
	%findall(D,xschm(_:_==>D),Decs),
	%findall(C,xschm(_:C==>_),Conds),
	findall(D,xtable(_,_,_,D),Decs),
	findall(C,xtable(_,_,C,_),Conds),
	flatten(Decs,FD),
	flatten(Conds,FC),
	subtract(FD,FC,GoalAtts).
	
get_goal_tables(GTables) :-
	find_goal_atts(GA),
	findall(I,list_goaltabs(GA,I),GTables).
	
list_goaltabs(GA,I) :-	
	%xschm(I:_==>DN),
	xtable(I,_,_,DN),
	subset(DN,GA).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Token-Driven-Firing

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Create stack, but fill also create database containg data about inputs, (num of tokens tables require)
% Number of tokens are determined by number of R->T liks (not T->T as previously)

runrules_TDI(Goals):-
	retractall(xtoken(_,_,_)),
	create_token_stack(Stack,Goals),
	fill_token_knowledge(Stack),
	run_token_inference(Stack).
	
fill_token_knowledge([]).
fill_token_knowledge([Head|Tail]) :-
	get_reqtok_number(Head,Req),
	assert(xtoken(Head,0,Req)),
	fill_token_knowledge(Tail),!.
	
% +Table -Number
get_reqtok_number(Table,Number) :-
	get_connections(Table,ConnSet),
	get_direct(ConnSet,Direct),
	length(Direct,Number).
	
get_direct([],[]).
get_direct([[S,D]|Rest],DirectRest) :-
	findall(Route,get_routes(S,D,Route),RoutesDuplicated),
	list_to_set(RoutesDuplicated,Routes),
	length(Routes,L),
	L > 1,
	get_direct(Rest,DirectRest).

get_direct([[S,D]|Rest],[[S,D]|DirectRest]) :-
	findall(Route,get_routes(S,D,Route),RoutesDuplicated),
	list_to_set(RoutesDuplicated,Routes),
	length(Routes,L),
	L = 1,
	get_direct(Rest,DirectRest).
	
get_connections(Table, ConnSet) :-
	findall([S,Table],(xrule S/_:_==>_:Table/_ ; 
		xrule S/_:_==>_**>_:Table/_ ; 
		xrule S/_:_==>_:Table ; 
		xrule S/_:_==>_**>_:Table),
		Connections),
	list_to_set(Connections,ConnSet).
	
get_routes(S,D,[S,D]) :-
	(xrule S/_:_==>_:T/_ ;  
	xrule S/_:_==>_**>_:T/_ ; 
	xrule S/_:_==>_:T ; 
	xrule S/_:_==>_**>_:T),
	D=T.
	
get_routes(S,D,[S|Route]) :-
	(xrule S/_:_==>_:T/_ ; 
		xrule S/_:_==>_**>_:T/_ ; 
		xrule S/_:_==>_:T ; 
		xrule S/_:_==>_**>_:T),
		D\=T,
		get_routes(T,D,Route).

create_token_stack(Stack,Goals) :-
	token_stack_helper(StackPartial,Goals),
	flatten(StackPartial,Stack).

token_stack_helper([],[]).
token_stack_helper([SH|Stack],[GH|Goals]) :-
	findall(Route,get_routes(_,GH,Route),RoutesDuplicated),
	list_to_set(RoutesDuplicated,Routes),
	union_paths(Routes,SH),
	token_stack_helper(Stack,Goals).
	

union_paths([PH|[]],PH).
union_paths([PH|Paths],Union) :-
	union_paths(Paths,PartialUnion),
	union(PH,PartialUnion,Union).


run_token_inference([]).
run_token_inference([Head|Stack]) :-
	runtable_TDI(Head),
	trajectory_projection(Head,table), 
	run_token_inference(Stack).

runtable_TDI(T) :-
	% Check if tokens are OK
	xtoken(T,Req,Req),
	%xschm(T:X==>Y),
	xtable(T,_,X,Y),
	heart_debug(1,['Firing tabl: ','(',T,')',X,'==>',Y]),
	runrules(T),
	nl,!.
runtable_TDI(_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DDI-Firing
%inference: DataDriven
%
%  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Goal driven for specified tables. Tables shoul be a list/set ex.: runrules_DDI([ms,dt,th]).
runrules_DDI(Init) :-
	create_ddi_stack(OrderedTables,Init),
	runtables(OrderedTables).
	
% Runs all tables in order they are linked
runrules_DDI :-
	init_tables(Init),
	create_ddi_stack(OrderedTables, Init),
	runtables(OrderedTables).

create_ddi_stack([],[]).
create_ddi_stack(Stack,[H|Goals]) :-
	create_ddi_stack_helper(StackPart1, H),
	create_ddi_stack(StackPart2,Goals),
	join_tables(StackPart1,StackPart2,StackDuplicated),
	reverse(StackDuplicated,RStackDuplicated),
	list_to_set(RStackDuplicated,RStack),
	reverse(RStack,Stack),!.
	
create_ddi_stack_helper(Stack,Table) :-
	get_succ_tables([Table],Successors),
	join_tables([Table],Successors,Stack),!.

get_succ_tables([],[]).
get_succ_tables(Tables,Successors) :-
	%findall(DecsAtts,((xschm Table : _ ==> DecsAtts),member(Table,Tables)),AllDecsAtts),
	findall(DecsAtts,(xtable(Table,_,_,DecsAtts),member(Table,Tables)),AllDecsAtts),
	flatten(AllDecsAtts,FlatDecsAtts),
	%findall(SucTable,((xschm SucTable : CondAtts ==> _), ord_intersect(FlatDecsAtts,CondAtts)),SucTables),
	findall(SucTable,(xtable(SucTable,_,CondAtts, _), ord_intersect(FlatDecsAtts,CondAtts)),SucTables),
	flatten(SucTables, SuccessorsPart1),		%??
	get_succ_tables(SuccessorsPart1,SuccessorsPart2),
	join_tables(SuccessorsPart1,SuccessorsPart2,Successors),!.
	

	
