%
% $Id: heart-io.pl,v 1.37 2009-12-26 16:43:39 esimon Exp $
%
%     Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
%     HeaRT the main shell
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2009   by S. Bobek, M. Gawedzki
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- write('HeaRT: $Id: heart-io.pl,v 1.37 2009-12-26 16:43:39 esimon Exp $'),nl.

:- dynamic(current_model/2).
:- dynamic(server_port/1).


%storage_dir('heart/storage').
storage_dir('./storage').
server_port_default(8090).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%% CALLBACKS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%all atts having 'comm: in'
findall_input_atts(IA) :-
	findall(N,(xattr(X),member(name:N,X),member(comm:in,X)),IA).

%fire all callbacks in given atts. Search for callback in att definition call it using Prolog Call
fireall_callbacks([Att|Rest]) :-
	fire_callback(Att),
	fireall_callbacks(Rest).

fireall_callbacks([]).

fire_callback(Att) :-
	xattr(A),
	member(name:Att,A),
	member(callback:[Name,Params],A),
	xcall Name : Params >>> Goal,
	call(Goal).

%all atts having 'comm: out'
findall_output_atts(OA) :-
	findall(N,(xattr(X),member(name:N,X),member(comm:out,X)),OA).


%Here goes interaction with outer world [DEBUG version]
%io_ask_att_console(Att) :-
%	write('Type value for attribute:  '),write(Att),nl,
%	read(Answer),
	%Add the value to the current state
%	alsv_new_val(Att,Answer).
	
% io_ask_att_console(MSG,Att) :-
	% write(MSG),nl,
	% read(Answer),
	% %Add the value to the current state
	% alsv_new_val(Att,Answer).

% io_ask_att_gui(Att) :-
	% jpl_new('JHeart.Therm',[],T),
	% alsv_domain(Att,[Min to Max],numeric),
	% concat_atom(['[',Att,',',Min,',',Max,']'],Parameter),
	% jpl_call(T,request,[T,Parameter],Answer),
	% atom_to_term(Answer,Answer2,_),
	% alsv_new_val(Att,Answer2).

% %Here goes communication with outer world [DEBUG version]
% io_tell_att_console(Att) :-
	% write('Attribute '),write(Att),write(' output value is '),
	% xstat(current:[Att,V]),
	% write(V),nl.
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%% TCP IO %%%%%%%%%%%%%%%%%%%%%%%%%%%%

srv :-
	server_port_default(Port),
	create_server(Port).

srv(Port) :-
	create_server(Port).

setup_storage :-
	storage_dir(Dir),
	exists_directory(Dir),
	access_file(Dir,write).
setup_storage :-
	storage_dir(Dir),
	make_directory(Dir),
	write('Storage directory for server: '),
	write(Dir), write(' created.').
setup_storage :-
	write('Cannot setup storage dir!'),
	!,
	fail.

create_server(Port) :-
    setup_storage,
    tcp_socket(Socket),
    tcp_bind(Socket, Port),
    tcp_listen(Socket, 1),
    tcp_open_socket(Socket, AcceptFd, _),
	heart_debug(2,['Server establisher on port ',Port]),!,
    dispatch(AcceptFd).


dispatch(AcceptFd) :-
    repeat,
	heart_debug(2,['Listening...']),
    tcp_accept(AcceptFd, Socket, _Peer),
    tcp_open_socket(Socket, In, Out),
	heart_debug(2,['Connection established']),
    handle_service(In, Out),
    close(In),
    close(Out),
	fail.
	

handle_service(In, Out) :-
	heart_debug(2,['Reading...']),
	style_check(-atom),
	set_stream(In,timeout(5)),
	catch(read_term(In,DataTerm,[]),_,fail),
	heart_debug(2,['Read from socket ',DataTerm]),
	handle(DataTerm,Response),
	heart_debug(2,['Response ',Response]),
	write_to_socket(Out,Response),!.
	
handle_service(_,Out):-
	write_to_socket(Out,[false,'Timeout. Bad or incomplete command.']),!.
	
handle([hello,_],[true,[hello,heart,1.0,5,[]]]).
	
handle([model,getlist], [true,Output]) :-
	heart_debug(2,['Handling request ','model getlist']),
	io_list_users(Users),
	io_list_users_models(Users,Output),!.	
	
handle([model,get,Modelname,Username], [true,Output]) :-
	heart_debug(2,['Handling request ','model get', 'for data',Modelname,Username]),
	catch(io_get_model(Username,Modelname, Output),_,fail),!.
handle([model,get,_,_],[false,'Model or username does not exist.']).	

handle([model,exists,Modelname,Usernme],[true,true]) :-
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',Usernme,'/',Modelname,'.pl'],FullPath),
	catch(exists_file(FullPath),_,fail),!.
handle([model,exists,Modelname,Usernme],[true,false]) :-
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',Usernme,'/',Modelname,'.pl'],FullPath),
	\+catch(exists_file(FullPath),_,fail),!.
handle([model,exists,_,_],[false,'Error while checkin file existance']).
	
handle([model,add,Modelname,Username, Data], Output) :-
	heart_debug(2,['Handling request ','model add', 'for data',Modelname,Username, Data]),
	io_save_model(Username,Modelname, Data),
	io_load_model(Username,Modelname),
	io_reload_current_model,
	Output = [true],!.
handle([model,add,_,_, _], [false,'Error while saving model.']).

handle([model,remove,Modelname,Username], Output) :-
	heart_debug(2,['Handling request ','model remove', 'for data',Modelname,Username]),
	io_delete_model(Username,Modelname),
	Output = [true],!.
handle([model,remove,_,_], [false,'Error while deleting model.']).
	
%State jest definicja stanu, np. [[day,workday],[hour,12]]
handle([model,run,Modelname,Username, Mode, StartTables, [HeadOf|StateValues]], [true,Output,Track]) :-		
	heart_debug(2,['Handling request ','model run (with state def)', 'for data',Modelname,Username, Mode, StartTables, [HeadOf|StateValues]]),
	catch(io_load_model(Username,Modelname),_,fail),
	state_add(current,[HeadOf|StateValues]),
	trajectory_set_mode(full),
	gox(current,StartTables,Mode),
	traj_id(TId),
	findall([Table,Rule],(xtraj(TId : FullTrajectory),member([Table/Rule,_],FullTrajectory)),TrackList),
	flatten(TrackList,Track),
	state_get_explicite(current,Output),!.
	
%state jest nazwa stanu, np. s1
handle([model,run,Modelname,Username, Mode, StartTables, StateName],[true,Output,Track]) :-		
	heart_debug(2,['Handling request ','model run (with statename)', 'for data',Modelname,Username, Mode, StartTables, StateName]),
	catch(io_load_model(Username,Modelname),_,fail),
	trajectory_set_mode(full),
	gox(StateName,StartTables,Mode),
	traj_id(TId),
	findall([Table,Rule],(xtraj(TId : FullTrajectory),member([Table/Rule,_],FullTrajectory)),TrackList),
	flatten(TrackList,Track),
	state_get_explicite(current,Output),!.
handle([model,run,_,_, _, _, _],[false,'\'Error occured while running model\'']).	

handle([state,add,Modelname,Username, Statename, [HeadOf|StateValues]], Output) :-		
	heart_debug(2,['Handling request ','state add ', 'for data',Modelname,Username, Statename, [HeadOf|StateValues]]),
	io_add_state(Username,Modelname, Statename, [HeadOf|StateValues]),
	Output = [true],!.
	
handle([state,add,_,_, _, _], [false,'Error while adding state']).

handle([state,add,Modelname,Username, Statename], Output) :-		
	heart_debug(2,['Handling request ','state remove', 'for data',Modelname,Username, Statename]),
	io_delete_state(Username,Modelname, Statename),
	Output = [true],!.
handle([state,delete,_,_, _, _], [false,'\'Error while deleting state\'']).


% Veryfication

handle([model,verify,vcomplete,Modelname,Username,Schema], [true,Output]) :-		
	heart_debug(2,['Handling request ','model verify vcomplete', 'for data',Modelname,Username,Schema]),
	io_load_model(Username,Modelname),
	vcomplete(Schema),
	get_current_vreport(VReport),
	Output = VReport,!.
	
handle([model,verify,vcomplete,_,_,_], [false,'Error while verifying completeness.']).
	
handle([model,verify,vcontradict,Modelname,Username,Schema], [true,Output]) :-		
	heart_debug(2,['Handling request ','model verify vcontradict', 'for data',Modelname,Username,Schema]),
	io_load_model(Username,Modelname),
	vcontradict(Schema),
	get_current_vreport(VReport),
	Output = VReport,!.

handle([model,verify,vcontradict,_,_,_], [false,'Error while verifying contradictions.']).
	
handle([model,verify,vreduce,Modelname,Username,Schema], [true,Output]) :-		
	heart_debug(2,['Handling request ','model verify vreduce', 'for data',Modelname,Username,Schema]),
	io_load_model(Username,Modelname),
	vreduce(Schema),
	get_current_vreport(VReport),
	Output = VReport,!.
handle([model,verify,vreduce,_,_,_], [false,'Error while verifying reductions.']).
	
	
handle([model,verify,vsubsume,Modelname,Username,Schema], [true,Output]) :-		
	heart_debug(2,['Handling request ','model verify vsubsume', 'for data',Modelname,Username,Schema]),
	io_load_model(Username,Modelname),
	vsubsume(Schema),
	get_current_vreport(VReport),
	Output = VReport,!.
handle([model,verify,vsubsume,_,_,_], [false,'Error while verifying subsumes.']).
	
handle(_,[false, 'Command not suported.']).

		
write_to_socket(Out,Msg) :-
	write(Out,Msg),
	nl(Out),
	flush_output(Out). 
	
% Saves model given as a list of characters (expression in quotation marks)
% Overrides existing models, Path is: Dir is USERNAME, Name is MODELNAME.pl
io_save_model(Dir, Name, Model) :-
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',Dir],Path),
	create_dir(Path),
	concat_atom([Path,'/',Name,'.pl'],FullPath),
	open(FullPath, write,File),
	write(File,Model), 
	close(File).
	
io_get_model(Dir,Name,Model) :-
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',Dir,'/',Name,'.pl'],FullPath),
	read_file_to_terms(FullPath,M1,[]),
	create_model_string(M1,M2),
	concat_atom(['[',M2,']'],Model).
	
io_delete_model(Dir,Name) :-
	current_model(Dir,Name),
	unload_current_model,
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',Dir,'/',Name,'.pl'],FullPath),
	delete_file(FullPath),!.

io_delete_model(Dir,Name) :-
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',Dir,'/',Name,'.pl'],FullPath),
	delete_file(FullPath).
	
io_load_model(Dir,Name) :-
	current_model(Dir,Name), !.
	
io_load_model(Dir,Name) :-
	save_current_model,
	unload_current_model,
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',Dir,'/',Name,'.pl'],FullPath),
	consult(FullPath),
	assert(current_model(Dir,Name)).
	
io_add_state(Dir,Name,Statename, Statevalues) :-
	io_load_model(Dir,Name),
	state_add(Statename,Statevalues).
	
io_delete_state(Dir,Name,Statename) :-
	io_load_model(Dir,Name),
	retractall(xstat(Statename:_)).
	
io_list_users(Users) :-
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/'],HeadOfPath),
	concat_atom([HeadOfPath,'*'],StoragePath),
	expand_file_name(StoragePath,UserPathList),
	findall(Username, (member(X,UserPathList),concat_atom([_,Username],HeadOfPath,X)),Users).
	
io_list_users_models([],[]).
io_list_users_models([User|Other],ModelsRecords) :-
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',User,'/'],HeadOfPath),
	concat_atom([HeadOfPath,'*'],StoragePath),
	expand_file_name(StoragePath,ModelsPathList),
		findall([User,Modelname],
		(member(X,ModelsPathList),concat_atom([_,ModelPL],HeadOfPath, X),concat_atom([Modelname,_],'.pl',ModelPL)),
		Models),
	io_list_users_models(Other,OtherModels),
	merge(Models,OtherModels,ModelsRecords).
	
unload_current_model :-
	retractall(xtype(_)),
	retractall(xtpgr(_)),
	retractall(xattr(_)),
	retractall(xatgr(_)),
	retractall(xschm(_)),
	retractall(xrule(_)),
	retractall(xstat(_)),
	retractall(xhalv(_)),
	retractall(xtraj(_)),
	retractall(current_model(_,_)).
	
io_reload_current_model :-
	current_model(Dir,Model),
	unload_current_model,
	io_load_model(Dir,Model).
	

save_current_model :-
	current_model(Dir,Name),
	storage_dir(StorageDir),
	concat_atom([StorageDir,'/',Dir,'/',Name,'.pl'],FullPath),
	tell(FullPath),
	% Begin telling
	listing([xtype,xtpgr,xattr,xatgr,xschm,xrule,xstat,xhalv,xtraj]),
	% End telling
	told.
	
save_current_model :-
	\+ current_model(_,_).

create_model_string([],'').	
create_model_string([Term|TRest],FullString) :-
	term_to_atom(Term,Atom),
	concat_atom([Atom,'.'],StringPart),
	create_model_string(TRest, SRest),
	concat_atom([StringPart,SRest],FullString).
	
	
create_dir(Path) :-
	exists_directory(Path),!.
	
create_dir(Path) :-
	make_directory(Path),!.
	

io_export_trajectory(Filename, TrajID) :-
	xtraj TrajID : Projection,
 	tell(Filename),
	io_make_traj_entries(Projection),
	told.

io_make_traj_entries([]).
io_make_traj_entries([[Rule, State]|Rest]) :-
	findall(V,(xstat State : V),Values),
	write('TRAJECTROY: Rule '),
	write(Rule), 
	write(' has generated state: '),
	write(Values),nl,
	io_make_traj_entries(Rest),!.
	

	
%[model,verify,vcomplete,user,model]
%[model,verify,vreduce,user,model]
%[model,verify,vcontradict,user,model]
%[model,verify,vsubsume,user,model]


%%%%%%%%%%%%% TEX Export %%%%%%%%%%%%%%%%%%%
io_export_TEX(Filename) :-
	create_model_TEX(TEXString),
	tell(Filename),
	write('\\documentclass[pdflatex,11pt]{article}'), nl,
	write('\\usepackage[polish]{babel}'), nl,
	write('\\usepackage[utf8]{inputenc}'), nl,
	write('\\usepackage[T1]{fontenc}'), nl,
	write('\\usepackage{times}'), nl,
	write('\\usepackage{amsmath}'), nl,
	write('\\begin{document}'), nl,
	write(TEXString),
	write('\\end{document}'),nl,
	told.
	
	
create_model_TEX(TEXString) :-
	create_types_TEX(TypesTEX),
	%create_typegrp_TEX(TypeGrpTEX),
	create_attrs_TEX(AttrsTEX),
	%create_attrgrp_TEX(AttrGrpTEX),
	%create_schms_TEX(SchmsTEX),
	create_rules_TEX(RulesTEX),
	%create_states_TEX(StatesTEX),
	concat_atom([TypesTEX,AttrsTEX,RulesTEX],TEXString),
	!.
	
create_types_TEX(TypesTEX) :-
	concat_atom(['\\begin{table}[h!]','\n',
	'\\caption{Types definitions}','\n',
	'\\begin{center}','\n',
	'\\begin{tabular}{|c||c|p{5cm}|c|c|p{5cm}|}','\n',
	'\\hline Name & Base & Domain & Scale & Ordered & Description \\\\ \\hline \\hline','\n'],TableBegin),
	findall(Type, xtype(Type),Types),
	create_types_TEX_helper(TableContent,Types),
	concat_atom(['\\end{tabular}','\n','\\end{center}','\n','\\end{table}','\n'],TableEnd),
	concat_atom([TableBegin,TableContent, TableEnd],TypesTEX),!.

create_types_TEX_helper('\\hline \n',[]).	
create_types_TEX_helper(TypesDefs,[Type|Types]) :-
	member(name:Name, Type),
	member(base:Base, Type),
	member(domain:Domain, Type),
	(member(scale:Scale, Type);Scale='$empty$'),
	(member(ordered:Ordered, Type);Ordered='$empty$'),
	(member(desc:Desc, Type);Desc='$empty$'),
	create_set_TEX(Domain,DomainText),
	create_types_TEX_helper(OtherTypesDefs,Types),
	concat_atom(['$',Name,'$&', Base,'&$', DomainText,'$&', Scale,'&', Ordered,'&', Desc, '\\\\ \\hline \n',OtherTypesDefs],TypesDefs).

create_set_TEX([Start to End], SetText) :-
	concat_atom(['\\left\\langle ',Start,';',End,'\\right\\rangle'],SetText),!.
	
create_set_TEX([Firts|SetDomain], SetText) :-
	create_set_TEX_helper([Firts|SetDomain],SetBody),
	concat_atom(['\\{',SetBody,'\\}'],SetText),!.

create_set_TEX_helper([Last|[]],TextLast):-
	term_to_atom(Last,TextLast).
create_set_TEX_helper([First|Rest],SetBodyText) :-
	create_set_TEX_helper(Rest,Elements),
	term_to_atom(First,TextFirst),
	concat_atom([TextFirst,',',Elements],SetBodyText).

create_attrs_TEX(AttsTEX) :-
	concat_atom(['\\begin{table}[h!]','\n',
	'\\caption{Attributes definitions}','\n',
	'\\begin{center}','\n',
	'\\begin{tabular}{|c||c|c|c|c|c|p{5cm}|}','\n',
	'\\hline Name & Calss & Type & Comm & Callback & Abbreviation & Description  \\\\ \\hline \\hline','\n'],TableBegin),
	findall(Att, xattr(Att),Atts),
	create_attrs_TEX_helper(TableContent,Atts),
	concat_atom(['\\end{tabular}','\n','\\end{center}','\n','\\end{table}','\n'],TableEnd),
	concat_atom([TableBegin,TableContent, TableEnd],AttsTEX),!.
	
create_attrs_TEX_helper('\\hline \n',[]).	
create_attrs_TEX_helper(AttrsDefs,[Attr|Attrs]) :-
	member(name:Name, Attr),
	member(class:Class, Attr),
	member(type:Type, Attr),
	member(comm:Comm, Attr),
	(member(callback:RawCallback, Attr),term_to_atom(RawCallback,Callback);Callback='empty'),
	(member(abbrev:Abbrev, Attr);Abbrev='empty'),
	(member(desc:Desc, Attr);Desc='$empty$'),
	create_attrs_TEX_helper(OtherAttrsDefs,Attrs),
	concat_atom(['$',Name,'$&', Class,'&$', Type,'$&', Comm,'&$', Callback,'$&$', Abbrev,'$&',Desc, '\\\\ \\hline \n',OtherAttrsDefs],AttrsDefs).

create_rules_TEX(RulesTEX) :-
	concat_atom(['\\begin{table}[h!]','\n',
	'\\caption{Rules definitions}','\n',
	'\\begin{center}','\n',
	'\\begin{tabular}{|c||p{5cm}|c|c|c|}','\n'],TableBegin),
	findall(SchemaName, (xschm SchemaName : _ ==> _),SchemaNames),
	create_rules_TEX_helper(SchemaNames, TableContent),
	concat_atom(['\\end{tabular}','\n','\\end{center}','\n','\\end{table}','\n'],TableEnd),
	concat_atom([TableBegin,TableContent, TableEnd],RulesTEX),!.

create_rules_TEX_helper([],'').
create_rules_TEX_helper([SchemaName|SchemaNames], TableContent) :-
	create_rows_TEX(SchemaName, SchemaRules),
	create_rules_TEX_helper(SchemaNames, RestTableContent),
	concat_atom(['\\multicolumn{5}{|c|}{Rules for schema ',SchemaName,'}\\\\ \\hline \n',SchemaRules,RestTableContent],TableContent).

create_rows_TEX(SchemaName, SchemaRules) :-
	findall(Rule, (xrule(Rule), Rule=(SchemaName/_ : _==>_)), Rules),
	create_rows_TEX_helper(Rules, Rows),
	concat_atom(['Rules ID & Conditions & Decisions & Actions & Links \\\\ \\hline \n',Rows],SchemaRules),!.
	
create_rows_TEX_helper([],'').
create_rows_TEX_helper([Rule|Rules], Rows) :- 
	((Rule = (_/ID : Conditions ==> Decisions **> [Action] : Link)) ;
	(Rule = (_/ID : Conditions ==> Decisions **> [Action] ),Link='$empty$') ;
	(Rule = (_/ID : Conditions ==> Decisions:Link),Action='$empty$') ;
	(Rule = (_/ID : Conditions ==> Decisions),Action='$empty$',Link='$empty$')),
	create_math_conditions(Conditions, MathConds),
	create_math_decisions(Decisions, MathDecs),
	create_rows_TEX_helper(Rules, OtherRows),
	term_to_atom(Link,TextLink),
	concat_atom([ID,'&$',MathConds,'$&$', MathDecs,'$&', Action,'&', TextLink, '\\\\ \\hline \n',OtherRows],Rows),!.
	
create_math_conditions([Cond|[]],TexCond) :- 
	alsv_to_TEX(Cond, TexCond).
create_math_conditions([Cond|Conditions], TexConds) :- 
	alsv_to_TEX(Cond, TexCond),
	create_math_conditions(Conditions, OtherTexConds),
	concat_atom([TexCond,'\\wedge ',OtherTexConds],TexConds),!.
	
alsv_to_TEX(Atribute eq [Head|Tail], TexCond) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '=', Set],TexCond),!.
	
alsv_to_TEX(Atribute neq [Head|Tail], TexCond) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '\\neq ', Set],TexCond),!.
	
alsv_to_TEX(Atribute subseteq [Head|Tail], TexCond) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '\\subseteq ', Set],TexCond),!.
	
alsv_to_TEX(Atribute supseteq [Head|Tail], TexCond) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '\\supseteq ', Set],TexCond),!.

alsv_to_TEX(Atribute sim [Head|Tail], TexCond) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '\\sim ', Set],TexCond),!.
	
alsv_to_TEX(Atribute notsim [Head|Tail], TexCond) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '\\not\\sim ', Set],TexCond),!.
	
alsv_to_TEX(Atribute in [Head|Tail], TexCond) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '\\in ', Set],TexCond),!.
	
alsv_to_TEX(Atribute notin [Head|Tail], TexCond) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '\\notin ', Set],TexCond),!.

alsv_to_TEX(Atribute eq AtomValue, TexCond) :-
	concat_atom([Atribute, '=',AtomValue],TexCond),!.
	
alsv_to_TEX(Atribute lt AtomValue, TexCond) :-
	concat_atom([Atribute, '<',AtomValue],TexCond),!.
alsv_to_TEX(Atribute lte AtomValue, TexCond) :-
	concat_atom([Atribute, '\\leq ',AtomValue],TexCond),!.
alsv_to_TEX(Atribute gt AtomValue, TexCond) :- 	
	concat_atom([Atribute, '>',AtomValue],TexCond),!.
alsv_to_TEX(Atribute gte AtomValue, TexCond) :-
	concat_atom([Atribute, '\\geq ',AtomValue],TexCond),!.
	
alsv_to_TEX(Atribute neq AtomValue, TexCond) :-
	concat_atom([Atribute, '\\neq ',AtomValue],TexCond),!.
	
% Decision and actions - have to be transformet like conditions - they are lists.

create_math_decisions([Atribute set [Head|Tail]|[]],TexDecs) :-
	create_set_TEX([Head|Tail], Set),
	concat_atom([Atribute, '= ',Set],TexDecs),!.
	
create_math_decisions([Atribute set [Head|Tail]|Rest], TexDecs) :-
	create_set_TEX([Head|Tail], Set),
	create_math_decisions(Rest,Result),
	concat_atom([Atribute, '= ',Set, ',' ,Result],TexDecs),!.
	
create_math_decisions([Atribute set Value|[]], TexDecs) :-
	term_to_atom(Value, AtomValue),
	concat_atom([Atribute, '= ',AtomValue],TexDecs),!.
	
create_math_decisions([Atribute set Value|Rest], TexDecs) :-
	term_to_atom(Value, AtomValue),
	create_math_decisions(Rest,Result),
	concat_atom([Atribute, '=',AtomValue,',',Result],TexDecs),!.
	

	
	
	
	
	
	
	
	
	
	
	
	
