

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%WTA%%%%%%%%%%%%%%%%%%%%%%%%%%%
% gox(StateIn,T,Mode,ExternalReasoner,InputFormat)                      %
%   ExternalReasoner - the one used to check the rules preconditions    %
%   InputFormat - e.g. RDF/XML (rdf) or HMR (hmr)                       %
%                                                                       %
% Augmented version of Heart - additional variable for an ext.reasoner  %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%WTA%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%WTA%%%%%%%%%%%%%%%%%%%%%%%%%%%
% gox(StateIn,T,Mode,ExternalReasoner)                                  %
%   ExternalReasoner - the one used to check the rules preconditions    %
%                                                                       %
% Augmented version of Heart - additional variable for an ext.reasoner  %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%WTA%%%%%%%%%%%%%%%%%%%%%%%%%%%

gox(StateIn,T,foi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runtables(T,Reasoner),
	trajectory_projection(end,simple), 
	state_print(current),!.

gox(StateIn,T,ddi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_DDI(T,Reasoner),
	trajectory_projection(end,simple), 
	state_print(current),!.

gox(StateIn,T,tdi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_TDI(T,Reasoner),
	trajectory_projection(end,simple), 
	state_print(current),!.
	
gox(StateIn,T,gdi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_GDI(T,Reasoner),
	trajectory_projection(end,simple), 
	state_print(current),!.

goxio(T,foi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runtables(T,Reasoner),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,ddi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_DDI(T,Reasoner),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,tdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_TDI(T,Reasoner),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,gdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_GDI(T,Reasoner),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Single Table processing                   %
%                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% table processing
runtables([],Reasoner).
runtables([R|T],Reasoner) :-
	runtable(R,Reasoner),
	trajectory_projection(R,table), 
	runtables(T,Reasoner).

runtable(T,Reasoner) :-
%	write('*Firing tabl: \"'),
%	write(T), write('\" ('),
	%FIXME: conditionless and decisionless rules support!!!
	%xschm(T:X==>Y),
	xtable(T,_,X,Y),
%	write(X),write('==>'),write(Y),write(')'),nl,
	heart_debug(1,['Firing tabl: ','(',T,')',X,'==>',Y]),
	runrules(T,Reasoner),
	nl.

runrules(T,Reasoner):-
	xrule(T/R:C==>D),
%	write(' *Firing rule: '), write(T/R),nl,
	heart_debug(1,['Firing rule: ',T/R]),
	cond_satisfy(C,Reasoner),
	process_decision(D,Go),
	inference_transfer(Go),
	trajectory_projection(T/R,full), 
	fail.
runrules(_,Reasoner).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GDI-Firing
%inference: GoalDriven GDI
%
%    * get set of goal tables
%    * build a sequential execution plan, by backward chaining using TT links, list of needed components
%    * run independent tables
%    * wave-like in steps
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Goal driven for specified tables. Tables shoul be a list/set ex.: runrules_GDI([os]).
runrules_GDI(Goals,Reasoner) :-
	create_stack(OrderedTables,Goals),
	write(OrderedTables),nl,
	runtables(OrderedTables,Reasoner).
	
% Runs all tables in order they are linked in fact - this implements DDI inference
runrules_GDI(Reasoner) :-
	get_goal_tables(Goals),
	write('Goals: '), write(Goals),nl,
	create_stack(OrderedTables, Goals),
	runtables(OrderedTables,Reasoner).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Token-Driven-Firing

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Create stack, but fill also create database containg data about inputs, (num of tokens tables require)
% Number of tokens are determined by number of R->T liks (not T->T as previously)

runrules_TDI(Goals,Reasoner):-
	retractall(xtoken(_,_,_)),
	create_token_stack(Stack,Goals),
	fill_token_knowledge(Stack),
	run_token_inference(Stack,Reasoner).

run_token_inference([],Reasoner).
run_token_inference([Head|Stack],Reasoner) :-
	runtable_TDI(Head,Reasoner),
	trajectory_projection(Head,table), 
	run_token_inference(Stack).

runtable_TDI(T,Reasoner) :-
	% Check if tokens are OK
	xtoken(T,Req,Req),
	%xschm(T:X==>Y),
	xtable(T,_,X,Y),
	heart_debug(1,['Firing tabl: ','(',T,')',X,'==>',Y]),
	runrules(T,Reasoner),
	nl,!.
runtable_TDI(_,Reasoner).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DDI-Firing
%inference: DataDriven
%
%  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Goal driven for specified tables. Tables shoul be a list/set ex.: runrules_DDI([ms,dt,th]).
runrules_DDI(Init,Reasoner) :-
	create_ddi_stack(OrderedTables,Init),
	runtables(OrderedTables,Reasoner).
	
% Runs all tables in order they are linked
runrules_DDI(Reasoner) :-
	init_tables(Init),
	create_ddi_stack(OrderedTables, Init),
	runtables(OrderedTables,Reasoner).






%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Reasoning with HeaRT-Pellet:                                      %
%  1. Build TBox: definitions of types and attributes               %
%    a) build additional statements: 'allDifferent' for individuals %
%  2. Call any inference mode you wish                              %
%  3. In each state build an ABox representing this state           %
%  4. Whenever you check a rule preconditions:                      %
%    a) build rule axioms (temporary TBox),                         %
%    b) ontology = definitions TBox + rule axioms TBox + state ABox %
%    c) send the ontology to Pellet to check its consistency        %
%  5. Interpret the result, carry on as usual                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Reasoning with HeaRT-Pellet - Alternative version:                %
%                                                                   %
%  1. Build TBox: definitions of types and attributes               %
%    a) build additional statements: 'allDifferent' for individuals   %
%  2. Call any inference mode you wish                              %
%  3. In each state build an ABox representing this state           %
%  4. Whenever you check a rule preconditions:                      %
%    a) build rule axioms (ABox statements)                         %
%    b) definitions TBox + rule axioms aBox + state ABox            %
%    c) send to Pellet to check realization the rule conditions     %
%  5. Interpret the result, carry on as usual                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Includes
:- ensure_loaded('heart-daal-translator.pl').
:- ensure_loaded('heart-pellet-comm.pl').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


gox(StateIn,T,foi,pellet) :-
	traj_genid(_),
        build_definition_TBox('daal/definitions.owl'),
        build_allDifferent('daal/allDifferent.owl'),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
        build_state_ABox(current,'daal/state.owl'),
	trajectory_projection(start,_), 
	runtables(T,pellet),
	trajectory_projection(end,simple), 
	state_print(current),!.


gox(StateIn,T,ddi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_DDI(T,Reasoner),
	trajectory_projection(end,simple), 
	state_print(current),!.

gox(StateIn,T,tdi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_TDI(T,Reasoner),
	trajectory_projection(end,simple), 
	state_print(current),!.
	
gox(StateIn,T,gdi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_GDI(T,Reasoner),
	trajectory_projection(end,simple), 
	state_print(current),!.

goxio(T,foi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runtables(T,Reasoner),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,ddi,Reasoner) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_DDI(T,Reasoner),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,tdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_TDI(T,Reasoner),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,gdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_GDI(T,Reasoner),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Single Table processing
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% table processing
runtables([],Reasoner).
runtables([R|T],Reasoner) :-
	runtable(R,Reasoner),
	trajectory_projection(R,table), 
	runtables(T,Reasoner).

runtable(T,Reasoner) :-
%	write('*Firing tabl: \"'),
%	write(T), write('\" ('),
	%FIXME: conditionless and decisionless rules support!!!
	%xschm(T:X==>Y),
	xtable(T,_,X,Y),
%	write(X),write('==>'),write(Y),write(')'),nl,
	heart_debug(1,['Firing tabl: ','(',T,')',X,'==>',Y]),
	runrules(T,Reasoner),
	nl.

runrules(T,Reasoner):-
	xrule(T/R:C==>D),
%	write(' *Firing rule: '), write(T/R),nl,
	heart_debug(1,['Firing rule: ',T/R]),
	cond_satisfy(C,Reasoner),
	process_decision(D,Go),
	inference_transfer(Go),
	trajectory_projection(T/R,full), 
	fail.
runrules(_,Reasoner).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GDI-Firing
%inference: GoalDriven GDI
%
%    * get set of goal tables
%    * build a sequential execution plan, by backward chaining using TT links, list of needed components
%    * run independent tables
%    * wave-like in steps
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Goal driven for specified tables. Tables shoul be a list/set ex.: runrules_GDI([os]).
runrules_GDI(Goals,Reasoner) :-
	create_stack(OrderedTables,Goals),
	write(OrderedTables),nl,
	runtables(OrderedTables,Reasoner).
	
% Runs all tables in order they are linked in fact - this implements DDI inference
runrules_GDI(Reasoner) :-
	get_goal_tables(Goals),
	write('Goals: '), write(Goals),nl,
	create_stack(OrderedTables, Goals),
	runtables(OrderedTables,Reasoner).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Token-Driven-Firing

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Create stack, but fill also create database containg data about inputs, (num of tokens tables require)
% Number of tokens are determined by number of R->T liks (not T->T as previously)

runrules_TDI(Goals,Reasoner):-
	retractall(xtoken(_,_,_)),
	create_token_stack(Stack,Goals),
	fill_token_knowledge(Stack),
	run_token_inference(Stack,Reasoner).

run_token_inference([],Reasoner).
run_token_inference([Head|Stack],Reasoner) :-
	runtable_TDI(Head,Reasoner),
	trajectory_projection(Head,table), 
	run_token_inference(Stack).

runtable_TDI(T,Reasoner) :-
	% Check if tokens are OK
	xtoken(T,Req,Req),
	%xschm(T:X==>Y),
	xtable(T,_,X,Y),
	heart_debug(1,['Firing tabl: ','(',T,')',X,'==>',Y]),
	runrules(T,Reasoner),
	nl,!.
runtable_TDI(_,Reasoner).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DDI-Firing
%inference: DataDriven
%
%  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Goal driven for specified tables. Tables shoul be a list/set ex.: runrules_DDI([ms,dt,th]).
runrules_DDI(Init,Reasoner) :-
	create_ddi_stack(OrderedTables,Init),
	runtables(OrderedTables,Reasoner).
	
% Runs all tables in order they are linked
runrules_DDI(Reasoner) :-
	init_tables(Init),
	create_ddi_stack(OrderedTables, Init),
	runtables(OrderedTables,Reasoner).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Rule in table processing

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% rule processing
%cond_satisfy([],Reasoner).
%cond_satisfy([C|Rest],Reasoner) :-
%	formula_satisfied(C,Reasoner),!,
%	cond_satisfy(Rest,Reasoner).


cond_satisfy(Conditions,pellet) :-
%Future: check if the condition has already been translated
    build_rule_axioms(Conditions,'daal/ruleAxioms.owl'),
    build_ontology(Ontology),
    check_consistency('ontology.owl',Reasoner),
    tell(pipe('rm daal/ruleAxioms.owl')).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%

build_ontology(Ontology):-
    tell('ontology.owl'),
    owl_xml_head,
    told,
    tell(pipe('for i in `ls daal`; do cat daal/$i>>ontology.owl; done')),
    told,
    append('ontology.owl'),
    owl_xml_tail,
    told.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_consistency(Ontology,pellet,Output):-
    pellet_command('consistency',Ontology,Output).
% sprawdz, czy w 1.linijce jest Consistent: Yes, jesli tak, cel jest spelniony.
%concat('sed ',Output, Check)
%tell(pipe(Check)).

check_consistency(Ontology,pellet):-
    pellet_command('consistency',Ontology).

check_consistency_detailed(Ontology,pellet):-
    pellet_command_detailed('consistency',Ontology).
    



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                           %
% DAAL translations.                        %
%  (predicates in heart-daal-translator.pl) %
%                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

build_definition_TBox(F):-
    tell(F),
    owl_xml_attr_gen,
    told.

build_allDifferent(F):-
    tell(F),
    owl_xml_allDifferent,
    told.
    	
% Argument: state id, e.g. 's1'
build_state_ABox(State,F):-
    tell(F),
    owl_xml_stat_gen(State),
    told.

% Argument: list of conditions, e.g. '[a1 eq v1, a2 in [v2,v3] ]'
build_rule_axioms(Cond,F):-
    tell(F),
    owl_xml_rule_cond_gen(Cond),
    told.

%% Argument: rule id, e.g. 'ms/1'
%build_rule_axioms(Rule,F):-
%    tell(F),
%    owl_xml_rulp_gen(Rule),
%    told.

