%
% $Id: heart-state.pl,v 1.10 2009-10-29 14:40:48 esimon Exp $
%
%     Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
%     HeaRT the main shell
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2009   by S. Bobek, M. Gawedzki
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- write('HeaRT: $Id: heart-state.pl,v 1.10 2009-10-29 14:40:48 esimon Exp $'),nl.

:- dynamic(stat_id/1).
:- dynamic(traj_id/1).
:- dynamic(trajectory_mode/1).

trajectory_mode(full). %can be full, table simple


state_copy(From,To) :-
	From \= To,
	retractall(xstat To:_),
	state_copy_helper(From,To).
	
state_copy(State,State).

state_copy_helper(From,To):-
	xstat From: Val,
	assert(xstat To: Val),
	fail.
	
state_copy_helper(_,_).
	
state_add(Statename, Statevalues) :-
	retractall(xstat Statename:_),
    state_add_helper(Statename,Statevalues).
	
state_add_helper(_,[]).
state_add_helper(Statename,[Value|Rest]) :-
	assert(xstat Statename:Value),
	state_add_helper(Statename,Rest).
	
state_get(Statename, Values) :-
	findall(Value,(xstat(Statename:Value)),Values).
	
state_get_explicite(Statename, Values) :-
	findall([Att,NormValue],(xstat(Statename:[Att,Value]),normalize(Att,Value,NormValue)),Values).
	

state_print(State) :-
	xstat State:_,
	write('State \"'), write(State), write('\": '),nl,
	state_print_helper(State),
	!.

state_print(State) :-
	heart_debug(1,['No state ',State,' defined.']).
	
state_print_helper(State):-
	xstat State:Val,
	write(Val),write(','),
	fail.
state_print_helper(_):-nl.

trajectory_projection(ProjectionName, Mode) :-
	% assert current state to new generated state
	trajectory_mode(Mode),
	stat_genid(SId),
	state_copy(current,SId),
	traj_id(TId),
	((xtraj TId : Trajectory);Trajectory = []),
	append([Trajectory,[[ProjectionName,SId]]],CurrentTraj),
	retractall(xtraj(TId:_)),
	assert(xtraj TId : CurrentTraj),
	!.
trajectory_projection(_,_).
	
trajectory_set_mode(Mode) :-
	retractall(trajectory_mode(_)),
	assert(trajectory_mode(Mode)).

stat_genid(ID) :-
	stat_id(ID),
	\+xstat(ID:_),!.

	
stat_genid(IDCandidate) :-
	repeat,
	gensym(s,IDCandidate),
	retractall(stat_id(_)),
	assert(stat_id(IDCandidate)),
	\+xstat(IDCandidate: _),!.
	

traj_genid(ID) :-
	traj_id(ID),
	\+xtraj(ID : _),!.
	
traj_genid(_) :-
	repeat,
	gensym(t,IDCandidate),
	retractall(traj_id(_)),
	assert(traj_id(IDCandidate)),
	\+xtraj(IDCandidate: _),!.
	
	
	
