assertState([],[]) :- !.
assertState([Attr| Attrs],[Value|Values]) :- 
  assertz(xstat vstate:[Attr,Value]),  assertState(Attrs,Values).

retractState([],[]) :- !.
retractState([Attr| Attrs],[Value|Values]) :- 
  retract(xstat vstate:[Attr,Value]),  retractState(Attrs,Values).

retractState :- retractall(xstat vstate:[_,_]).

writeState([],[],[]) :- !.
writeState([C|Cs],[V|Vs],[[C,ExpandedV]|List]) :- 
  expand(C,V,ExpandedV),
  writeState(Cs,Vs,List).

expand(_,[],[]) :- !.  
expand(C,[H1|V],[H2|EV]) :- 
	vnormalize(C,H1,H2),!,
	expand(C,V,EV).
expand(C,V,EV) :-
	vnormalize(C,V,EV),!.


vappend(L,[],[L]):-!.
vappend([],L,L):-!.
vappend(Elem, List, [Elem|List]):-!.

appendList([], L, L) :- !.
appendList([Elem|L1], L2, [Elem|L3]) :- appendList(L1, L2, L3).

appendDistinct([], L, L) :- !.
appendDistinct([Elem|L1], L2, [Elem|L3]) :- 
  \+member(Elem,L2), appendDistinct(L1, L2, L3),!.
appendDistinct([Elem|L1], L2, L3) :- 
  member(Elem,L2), appendDistinct(L1, L2, L3).

lastElem(Elem, List) :- appendList(_,[Elem],List).

next(X, Y) :- number(Y), X is Y - 1,!. 
next(X, Y) :- number(X), Y is X + 1,!.

%vget_covered_states(Schema,Att, Domain) :-
%	findall(any,(xrule Schema/_: F==>_,(
%		member(Att eq any,F);
%		member(Att in any,F))),
%		[_|_]),
%	alsv_domain(Att,Domain,_),!.

vget_covered_states(Schema,Att,FinalStates) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,Domain, _),
	normalize(Att,Domain, NDomain),
	findall(S, (xrule Schema/_: F==>_,(
		(member(Att in S,F),S \= any, S \= [any]);
		((member(Att in any,F);member(Att in [any],F)),S=NDomain);
		(member(Att eq S,F),S \= any);
		(member(Att eq any,F),S=NDomain);
		member(Att neq S,F);
		member(Att notin S,F);
		(member(Att lt SD,F),normalize(Att,SD,NSD),vminimum(NDomain,L),U is NSD-1,S=[L to U], L < U);
		(member(Att gt SD,F),normalize(Att,SD,NSD),vmaximum(NDomain,U),L is NSD+1,S=[L to U], L < U);
		(member(Att lte SD,F),normalize(Att,SD,NSD),vminimum(NDomain,L),S=[L to NSD]);
		(member(Att gte SD,F),normalize(Att,SD,NSD),vmaximum(NDomain,U),S=[NSD to U]))
	),
	States),
	flatten(States,DupStates),
	list_to_set(DupStates,FinalStates),!.

vget_covered_states(Schema,Att,FinalStates) :-
	alsv_attr_class(Att,general),
	alsv_domain(Att,Domain, _),
	normalize(Att,Domain, NDomain),
	findall(S, (xrule Schema/_: F==>_,(
		(member(Att eq S,F),S \= any, S \=[any]);
		((member(Att eq any,F);member(Att eq [any],F)),S=NDomain);
		member(Att neq S,F);
		member(Att subseteq S,F);
		member(Att supseteq S,F);
		member(Att sim S,F);
		member(Att notsim S,F))
	),
	States),
	flatten(States,DupStates),
	list_to_set(DupStates,FinalStates),!.


vget_rules_states(Schema, FinalStates) :-
  findall([ID,Cond],(xrule Schema/ID: Cond ==> _),Conds),
  vget_rules_states_helper(Conds, FinalStates).


vget_rules_states_helper([],[]).
vget_rules_states_helper([[ID,C]|Conds], [[ID,RuleState]|States]) :-
  findall([Att,S],(
  		alsv_attr_class(Att,simple),
                alsv_domain(Att,Domain,_),
  		normalize(Att,Domain,NDomain),
		(member(Att in S,C);
		(member(Att eq S,C),S \= any);
		(member(Att eq any,C),S=NDomain);
		member(Att neq S,C);
		member(Att notin S,C);
		(member(Att lt SD,C),normalize(Att,SD,NSD),vminimum(NDomain,L),U is NSD-1,S=[L to U],L < U);
		(member(Att gt SD,C),normalize(Att,SD,NSD),vmaximum(NDomain,U),L is NSD+1,S=[L to U]), L < U;
		(member(Att lte SD,C),normalize(Att,SD,NSD),vminimum(NDomain,L),S=[L to NSD]);
		(member(Att gte SD,C),normalize(Att,SD,NSD),vmaximum(NDomain,U),S=[NSD to U]))
	),
        RuleState),!,
	vget_rules_states_helper(Conds,States).

vget_rules_states_helper([[ID,C]|Conds], [[ID,RuleState]|States]) :-
  findall([Att,S],(		
  		alsv_attr_class(Att,general),
		alsv_domain(Att,Domain,_),
		((member(Att eq S,C),S \= any);
		(member(Att eq any,C),S=Domain);
		member(Att neq S,C);
		member(Att subseteq S,C);
		member(Att supseteq S,C);
		member(Att sim S,C);
		member(Att notsim S,C))
	),
        RuleState),!,
	vget_rules_states_helper(Conds,States).
        


vmaximum([],end):-!.

vmaximum([_ to U|Set], InnerMaximum) :-
	vmaximum(Set,InnerMaximum),
	InnerMaximum\=end,
	InnerMaximum > U,!.

vmaximum([_ to U|_], U):-!.
	
vmaximum([H|Set], InnerMaximum) :-
	vmaximum(Set,InnerMaximum),
	InnerMaximum\=end,
	InnerMaximum > H,!.

vmaximum([H|_], H) :- !.	 

vminimum([],end):-!.

vminimum([L to _|Set], InnerMinimum) :-
	vminimum(Set,InnerMinimum),
	InnerMinimum\=end,
	InnerMinimum < L,!.

vminimum([L to _|_], L) :- !.
	
vminimum([H|Set], InnerMinimum) :-
	vminimum(Set,InnerMinimum),
	InnerMinimum\=end,
	InnerMinimum < H,!.

vminimum([H|_], H) :- !.	 
	

vcondition(Att eq any, Att in Domain) :- 
  alsv_domain(Att,Domain,_), !.
vcondition(Att lt Value, Att in [LB to UB]) :- 
  alsv_domain(Att,Domain,numeric),
  vnormalize(Att, Domain, [LB|_]),
  vnormalize(Att, [Value], [V]),
  next(UB, V), !.
vcondition(Att gt Value, Att in [LB to UB]) :- 
  alsv_domain(Att,Domain,numeric),
  vnormalize(Att, Domain, NDomain),
  lastElem(UB,NDomain),
  vnormalize(Att, [Value], [V]),  
  next(V,LB), !.
vcondition(C,C).
 
%vnormalize(_,[LB to LB], [FLB]) :- FLB is LB*1.0,!.
%vnormalize(_,[LB to UB], Output) :- 
%	next(LB, P), 
%	P =< UB, 
%	vnormalize(_,[P to UB], Output2), 
%	FLB is LB*1.0,
%	append(FLB, Output2, Output),!.  
vnormalize(Att, Value, Output) :- normalize(Att, Value, Output), !.

vcomplement(Attr, V, VC):-
  vnormalize(Attr, V, NV),
  alsv_domain(Attr, Domain, _),
  vnormalize(Attr, Domain, NDomain),
  subtract(NDomain,NV, VC).

split_decision(D**>A:T/R,[D,A,T,R]) :- !.
split_decision(D**>A:T,[D,A,T,vnone]) :- !.
split_decision(D**>A,[D,A,vnone,vnone]) :- !.
split_decision(D:T/R,[D,[],T,R]) :- !.
split_decision(D:T,[D,[],T,vnone]) :- !.
split_decision(D,[D,[],vnone,vnone]) :- !.

vreport(N,D) :-
	flag_vreport(L),
	vid(ID),
	assert(xhalv(ID: D)),
	N =< L,
	vreport_process(D),!.
vreport(_,_).

vreport_process([]) :- nl,!.
vreport_process([H|T]):-
	write(H), write(' '),!,
	vreport_process(T).
	
get_current_vreport(VReport) :-
	vid(ID),
	findall(R,xhalv(ID:R),VReport),!.

get_current_vreport([]).

	
vgenid(ID) :-
	vid(ID),
	\+xhalv(ID : _),!.
	
vgenid(_) :-
	repeat,
	gensym(v,IDCandidate),
	retractall(vid(_)),
	assert(vid(IDCandidate)),
	\+xhalv(IDCandidate: _),!.

comb(0,_,[]).
comb(N,[X|T],[X|Comb]):-
	N>0,N1 is N-1,comb(N1,T,Comb),
	\+member([X,_],Comb).
comb(N,[_|T],Comb):-N>0,comb(N,T,Comb).

vdistinct_states_comb(Schema, List, Comb) :-
	xtable(Schema, _,From,_),
	length(From,N),
        findall(OK, (member(OK,List),OK=[_,[_|_]]),OKs),
	findall(C,(comb(N,OKs,C),
		findall(F,member([F,_],C),M),
		list_to_set(M,M)),
		Comb),!.


vconcat([String],String) :- !.
vconcat([E1,E2|List], String) :- concat(E1,E2,Elem), vconcat([Elem|List], String).

vconcat_ws(_, [String], String) :- !.
vconcat_ws(Separator, [E1,E2|List], String):- vconcat([E1,Separator,E2], Elem),vconcat_ws(Separator,[Elem|List], String).
