%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% INFERENCE FOR ATOMIC FORMULAE                               %%%
%%% class : simple                                              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%-- eq -----------------------------------------------------------%
valid(Attr eq Value) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,[Value],[NValue]),
  v(Attr eq Value2),
  vnormalize(Attr,[Value2],[NValue]), !.
  
valid(Attr eq Value) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,[Value],[NValue1]), 
  v(Attr neq Value2),
  vnormalize(Attr,[Value2],[NValue2]), 
  NValue1 \= NValue2, !.
  
valid(Attr eq Value) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,[Value],[NValue]),
  v(Attr in ValueSet), 
  vnormalize(Attr,ValueSet,NValueSet),
  member(NValue, NValueSet), !.
  
valid(Attr eq Value) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,[Value],[NValue]), 
  v(Attr notin ValueSet), 
  vnormalize(Attr,ValueSet,NValueSet),
  \+member(NValue, NValueSet), !.

%-- neq ---------------------------------------------------------%
valid(Attr neq Value) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,[Value],[NValue]), 
  v(Attr neq Value2),
  vnormalize(Attr,[Value2],[NValue2]),
  NValue = NValue2, !.
  
valid(Attr neq Value) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,[Value],[NValue]), 
  alsv_domain(Attr,Domain,_), 
  vnormalize(Attr,Domain,NDomain), 
  subtract(NDomain, [NValue],NValueSet1), 
  v(Attr in ValueSet2), 
  vnormalize(Attr,ValueSet2, NValueSet1), !.
  
valid(Attr neq Value) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,[Value],[NValue]), 
  v(Attr notin ValueSet),
  vnormalize(Attr,ValueSet,[NValue]),!.

%-- in -----------------------------------------------------------%
valid(Attr in ValueSet) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,ValueSet,NValueSet),
  v(Attr eq Value),
  vnormalize(Attr,[Value],NValueSet), !.
  
valid(Attr in ValueSet) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,ValueSet,NValueSet),
  v(Attr neq Value), 
  vnormalize(Attr,[Value],NValue),
  \+member(NValue,NValueSet), !.
  
valid(Attr in ValueSet) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,ValueSet,NValueSet),
  v(Attr in ValueSet2), 
  vnormalize(Attr,ValueSet2,NValueSet2), 
  subset(NValueSet, NValueSet2), !.
  
valid(Attr in ValueSet) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,ValueSet,NValueSet),
  v(Attr notin ValueSet2), 
  vnormalize(Attr,ValueSet2,NValueSet2),
  intersection(NValueSet, NValueSet2, []), !.

%-- notin --------------------------------------------------------%
valid(Attr notin ValueSet) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,ValueSet,NValueSet),
  alsv_domain(Attr,Domain,_), 
  vnormalize(Attr,Domain,NDomain), 
  subtract(NDomain, NValueSet, NValueSet2), 
  v(Attr eq Value),
  vnormalize(Attr,[Value], NValueSet2), !.
  
valid(Attr notin ValueSet) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,ValueSet,NValueSet),
  v(Attr neq Value), 
  vnormalize(Attr,[Value], NValueSet), !.
  
valid(Attr notin ValueSet) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,ValueSet,NValueSet),
  alsv_domain(Attr,Domain,_), 
  vnormalize(Attr,Domain,NDomain), 
  subtract(NDomain, NValueSet, NValueSet1), 
  v(Attr in ValueSet2),
  vnormalize(Attr, ValueSet2, NValueSet1), !.
  
valid(Attr notin ValueSet) :- 
  alsv_attr_class(Attr,simple), 
  vnormalize(Attr,ValueSet,NValueSet),
  v(Attr notin ValueSet2), 
  vnormalize(Attr, ValueSet2, NValueSet2),
  subset(NValueSet2, NValueSet), !. 
  
  
  
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% INFERENCE FOR ATOMIC FORMULAE                               %%%
%%% class : general                                             %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%-- eq -----------------------------------------------------------%
valid(Attr eq VSet) :- 
  alsv_attr_class(Attr,general), 
  vnormalize(Attr,VSet,NVSet),
  v(Attr eq WSet),
  vnormalize(Attr,WSet,NVSet), !.
  
valid(Attr eq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr neq WSet), 
  vnormalize(Attr,WSet,NWSet),
  neqset(NVSet,NWSet), !.
  
valid(Attr eq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr subseteq WSet), 
  vnormalize(Attr,WSet,NWSet),
  subset(NVSet, NWSet), !.
  
valid(Attr eq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr supseteq WSet), 
  vnormalize(Attr,WSet,NWSet),
  subset(NWSet, NVSet), !.

valid(Attr eq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr sim WSet), 
  vnormalize(Attr,WSet,NWSet),
  intersection(NVSet, NWSet,[_|_]), !.
  
valid(Attr eq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr notsim WSet), 
  vnormalize(Attr,WSet,NWSet), 
  intersection(NVSet, NWSet,[]), !.

%-- neq ----------------------------------------------------------% 
valid(Attr neq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr neq WSet), 
  vnormalize(Attr,WSet,NVSet), !.
  
valid(Attr neq _) :- 
  alsv_attr_class(Attr,general),
  %vnormalize(Attr,VSet,NVSet),
  alsv_domain(Attr,Domain,_),
  vnormalize(Attr, Domain, NDomain),
  v(Attr subseteq WSet), 
  vnormalize(Attr, WSet, NDomain), !.

valid(Attr neq _) :- 
  alsv_attr_class(Attr,general),
  %vnormalize(Attr,VSet,NVSet),
  alsv_domain(Attr,Domain,_),
  vnormalize(Attr,Domain,NDomain),
  v(Attr sim WSet), 
  vnormalize(Attr, WSet, NDomain), !.
  
%-- subseteq -----------------------------------------------------%
valid(Attr subseteq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr neq WSet), 
  vnormalize(Attr, WSet, NWSet),
  subset(NVSet, NWSet), 
  neqset(NVSet,NWSet), !.
  
valid(Attr subseteq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr subseteq WSet), 
  vnormalize(Attr, WSet, NWSet),
  subset(NVSet, NWSet), !.

valid(Attr subseteq _) :- 
  alsv_attr_class(Attr,general),
  %vnormalize(Attr,VSet,NVSet),
  alsv_domain(Attr,Domain,_),
  vnormalize(Attr,Domain,NDomain),
  v(Attr sim WSet), 
  vnormalize(Attr, WSet, NDomain), !.
  
valid(Attr subseteq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr notsim WSet), 
  vnormalize(Attr, WSet, NWSet),
  intersection(NVSet, NWSet,[]), !.

%-- subseteq -----------------------------------------------------% 
valid(Attr supseteq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr neq WSet), 
  vnormalize(Attr, WSet, NWSet),
  subset(NWSet, NVSet), 
  neqset(NVSet,NWSet), !.
  
valid(Attr supseteq _) :- 
  alsv_attr_class(Attr,general),
  %vnormalize(Attr,VSet,NVSet),
  alsv_domain(Attr,Domain,_),
  vnormalize(Attr,Domain,NDomain),
  v(Attr subseteq WSet), 
  vnormalize(Attr, WSet, NDomain),!.
  
valid(Attr supseteq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr supseteq WSet), 
  vnormalize(Attr, WSet, NWSet),
  subset(NWSet, NVSet), !.

valid(Attr supseteq VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr sim WSet), 
  vnormalize(Attr, WSet, NWSet),
  intersection(NVSet, NWSet,[_|_]), !.
  
%-- eq -----------------------------------------------------------%
valid(Attr sim VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr neq WSet), 
  vnormalize(Attr, WSet, NWSet),
  intersection(NVSet, NWSet,[]), !.
  
valid(Attr sim _) :- 
  alsv_attr_class(Attr,general),  
  %vnormalize(Attr,VSet,NVSet),
  alsv_domain(Attr,Domain,_),
  vnormalize(Attr,Domain,NDomain),
  v(Attr subseteq WSet), 
  vnormalize(Attr, WSet, NDomain), !.

valid(Attr sim VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr sim WSet), 
  vnormalize(Attr, WSet, NVSet), !.
  
%-- notsim -------------------------------------------------------%
valid(Attr notsim VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr neq WSet), 
  vnormalize(Attr, WSet, NWSet),
  intersection(NVSet, NWSet,[_|_]), !.
  
valid(Attr notsim _) :- 
  alsv_attr_class(Attr,general),
  %vnormalize(Attr,VSet,NVSet),
  alsv_domain(Attr,Domain,_),
  vnormalize(Attr,Domain,NDomain),
  v(Attr subseteq WSet), 
  vnormalize(Attr, WSet, NDomain), !.

valid(Attr notsim _) :- 
  %vnormalize(Attr,VSet,NVSet),
  alsv_domain(Attr,Domain,_),
  vnormalize(Attr,Domain,NDomain),
  v(Attr sim WSet), 
  vnormalize(Attr, WSet, NDomain), !.
  
valid(Attr notsim VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr notsim WSet), 
  vnormalize(Attr, WSet, NVSet), !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%  valid decision part of rule   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
valid(Attr set Value) :- 
  alsv_attr_class(Attr,simple),
  vnormalize(Attr, [Value], [NValue]),
  v(Attr set Value2), 
  vnormalize(Attr, [Value2], [NValue]), !.

valid(Attr set VSet) :- 
  alsv_attr_class(Attr,general),
  vnormalize(Attr,VSet,NVSet),
  v(Attr set WSet), 
  vnormalize(Attr, WSet, NVSet), !.


