:- dynamic(v/1).
:- dynamic(j/1).
:- dynamic(c/1).
:- dynamic(vid/1).
:- dynamic(xhalv/1).

:- op(830,fx,xhalv).

:- ensure_loaded('util.pl').
:- ensure_loaded('valid.pl').
:- ensure_loaded('invalid.pl').
:- ensure_loaded('join.pl').

flag_vreport(4).

:- add_nb_set(vgenid/1,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% COMPLETENESS                                                             %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- MISSING RULES ------------------------------------------------------------%
%------------------------------------------------------------------------------%	

vcomplete(Schema) :- 
  vgenid(_),		% Non-backtrackable predicate
  xtable(Schema, _, Attrs, _),
  vget_states(Schema,Attrs,UStates,CStates),
  %flatten(NestUStates,UStates),
  vsplit2details(CStates,DCStates),
  vsplit2details(UStates,DUStates),
  merge(DUStates,DCStates,States),
  vsplit_intersections(States,DisjointStates),
  vdistinct_states_comb(Schema, DisjointStates, Combinations),
  %rules states
  vstate_cover(Schema, Combinations),!,
  retractall(c(_)).
%vcomplete(_) :- 
 % findall(C,c(C),[]),!.
%vcomplete(_) :-  
%  retractall(c(_)),!.

  
vget_states(_,[],[],[]).
vget_states(Schema,[A|ARest], UStates, CStates) :- 
  alsv_domain(A, Domain,_), write('State: '),write(A),nl,
  normalize(A,Domain, NDomain),
  vget_covered_states(Schema,A,States), %cominations of existing states
  normalize(A,States,NState),
  sub(NDomain, NState,UState1),
  vget_states(Schema,ARest, UStates2,CStates2),
  merge([[A,NState]],CStates2, CStates),
  merge([[A,UState1]],UStates2,UStates).

vsplit2details([],[]).
vsplit2details([[Att,Val]|States],Result) :-
  vsplit2details_helper(Att,Val, SingleRes),
  vsplit2details(States,OtherRes),
  merge(SingleRes,OtherRes,Result).

vsplit2details_helper(_,[],[]).
vsplit2details_helper(Att,[V|Vals],[[Att,[V]]|SingleRes]) :-
  vsplit2details_helper(Att,Vals,SingleRes).

vsplit_intersections([],[]).
vsplit_intersections([[A,V]|States], DisjointStates) :-
  findall(Vs,member([A,Vs],[[A,V]|States]),List),
  flatten(List,FList),
  vsplit_intersections_helper(FList, DisjointList),
  findall([A,[DV]],member(DV,DisjointList),DisStates1),
  findall([AA,VV],(member([AA,VV],States), AA \= A), RestStates),
  vsplit_intersections(RestStates,DisStates2),
  merge(DisStates1,DisStates2,DisjointStates).


vsplit_intersections_helper([],[]).
vsplit_intersections_helper([V|List], DisjointList) :-
  alsv_intersect([V],List, R),
  R \= [],
  sub(List,R, Result),
  merge(R,Result, NewList),
  vsplit_intersections_helper(NewList,DisjointList),!.

vsplit_intersections_helper([V|NewList], [V|DisjointList]) :-
  vsplit_intersections_helper(NewList,DisjointList),!.
  

%try to fire rules from schema. If any is OK, than state is covered.
vis_covered([[_,CState]|_],StateSet) :-
%  write(CState),write(' '),nl,write(StateSet),
  vis_covered_helper(CState, StateSet),!. % if doeasn't fail, means that the CState rule will be fired for the following StateSet.
%  write(' COVERED!'),nl,nl,!.
%  vis_covered(Covered,StateSet), !.

vis_covered([[_,_]|Covered],StateSet) :-
  Covered \= [],
 % write(' Not COvered...'),nl,nl,
  vis_covered(Covered,StateSet),!.


vis_covered_helper([],_).
vis_covered_helper([[A,V]|Covered],StateSet) :-
  flatten([V], CoveredValue),
  normalize(A,CoveredValue,NCV),	
  member([A,SV],StateSet),
  alsv_subset(SV,NCV),
  vis_covered_helper(Covered,StateSet).

  
vstate_cover(_,[]).
vstate_cover(Schema, [StateSet|R]) :-
  vget_rules_states(Schema,CoveredStates),
  vis_covered(CoveredStates,StateSet),
  vstate_cover(Schema,R),!.

vstate_cover(Schema, [StateSet|R]) :-
  vcover_helper(Schema, StateSet),
  vstate_cover(Schema, R),!.
  
vcover_helper(Schema, StateSet) :- 
  vcover(Schema, StateSet, Res),
  term_to_atom(Res, AtomRes),
  vconcat(['\'In ', Schema,' uncover state: ', AtomRes ,'\''],RString), 
  vreport(1,[RString]), nl .

vcover(_,[],'').
vcover(Schema, [[_, []]|R],Res) :-
  vcover(Schema,R,Res).
vcover(Schema, [[Attr,State]|R],Res) :-
  vcover(Schema, R, InnerRes),
  term_to_atom(State,AState),
  concat_atom(['(',Attr,') in (',AState,')',InnerRes],Res),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate([],[]).
generate([A|ARest], State) :- 
  alsv_domain(A, Domain,_), 
  vnormalize(A,Domain, NDomain), 
  member(V,NDomain),			
  generate(ARest, AState), 
  vappend(V, AState, State).

covered([]) :- !.
covered([Fact|Facts]) :- 
  vcondition(Fact,VFact),
  alsv_valid(VFact,vstate), 
  covered(Facts).
  
generateRule([],[]) :- !.
generateRule([A|Attrs],[V|Values]) :- 
  c(A in ValSet),
  flatten([V], NV),
  appendDistinct(NV, ValSet,NewValSet),
  retract(c(A in ValSet)),
  assert(c(A in NewValSet)),
  generateRule(Attrs,Values),!.
generateRule([A|Attrs],[V|Values]) :-
  flatten([V], FV),
  assert(c(A in FV)),
  generateRule(Attrs,Values). 
   
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% DETERMINISM                                                              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- CONTRADICTION ------------------------------------------------------------%
%------------------------------------------------------------------------------%
vcontradict(Schema) :- 
  vgenid(_), % Non-backtrackable predicate
  %xschm Schema: Attrs ==> _, 
  %generate(Attrs,Values), 
  vget_rules_states(Schema, RuleStates),
  contr(Schema, RuleStates),
  fail.
vcontradict(_). 

% find all rules states, and test if any of them intersects with other 
% the intersection is the colision set. (if "decision part is different)

contr(_,[]).
contr(Schema,[[ID, CStates]|Rest])  :-
  contr_helper(Schema, ID, CStates, Rest, Conflicts),
  report_conflicts(Schema,Conflicts),
  contr(Schema,Rest),!.

contr_helper(_,_,_,[],[]).
contr_helper(Schema, ID1, SingleState, [[ID2, CStates]|AllRest],[[ID1, ID2, State]|C]) :-
  contr_helper_helper(Schema, ID1, ID2, SingleState, CStates, State),
  xrule(Schema/ID1: _==>Dec1),
  xrule(Schema/ID2: _==>Dec2),
  conflict(Dec1,Dec2),
  contr_helper(Schema, ID1, SingleState, AllRest, C),!.

contr_helper(Schema, ID1, SingleState, [_|AllRest],C) :-
  contr_helper(Schema, ID1, SingleState, AllRest, C),!.
  

contr_helper_helper(_, _, _, [],_,[]).
contr_helper_helper(Schema, ID1, ID2, [[Att,V1]|StateRest], CStates, [[Att,Confl]|ConflRest]) :-
  member([Att, V2],CStates),
  flatten([V1],FV1),
  flatten([V2], FV2),
  normalize(Att,FV2, NFV2),
  normalize(Att,FV1, NFV1),
  alsv_intersect(NFV1,NFV2,Confl), 
  Confl \= [],
  contr_helper_helper(Schema, ID1, ID2, StateRest, CStates, ConflRest).

report_conflicts(_,[]).
report_conflicts(Schema,[[ID1, ID2, State]|Rest]) :-
  term_to_atom(State,AState),
  vconcat(['\'In ', Schema,' rule: ',ID1,' conflicts with rule ',ID2,' for state: ', AState, '\''],RString), 
  vreport(0,[ID1,RString]),
  %vreport(0,[ID2,RString]),
  report_conflicts(Schema,Rest),!.
  

%contr(Schema, Attrs, Values) :- 
  %assertState(Attrs,Values),
  %xrule(Schema/Id: Facts ==> Decs), 
  %xrule(Schema/Id2: Facts2 ==> Decs2),
  %Id \= Id2, 
  %covered(Facts), %vis_covered(CoveredStates
  %covered(Facts2), 
  %conflict(Decs, Decs2),
  %retractState(Attrs,Values),
  %vconcat_ws(', ', Attrs, AString),
  %vconcat_ws(', ', Values, VString),
  %vconcat(['\'In ', Schema,' rule: ',Id,' conflicts with rule ',Id2,' for state: (',AString,') = (',VString,')\''],RString), 
  %vreport(0,[Id,RString]),
  %vreport(0,[Id2,RString]),!.
%contr(_, Attrs, Values) :- 
  %retractState(Attrs,Values).
  
conflict(Decs1, Decs2):-
  split_decision(Decs1,[D1,_,_,_]), 
  split_decision(Decs2,[D2,_,_,_]),
  conflict_dec_attributes(D1,D2).
  
conflict_dec_attributes([],[]) :- !.
conflict_dec_attributes([D1|Ds1],[D2|Ds2]) :-  
  conflict_item(D1,D2),
  conflict_dec_attributes(Ds1,Ds2).  

conflict_item(D1,D2) :-
  retractall(v(_)),
  assert(v(D2)), 
  invalid(D1).

%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% MIN. NUMBER OF RULES                                                     %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- SUBSUMPTION --------------------------------------------------------------%
%------------------------------------------------------------------------------%
vsubsume(Schema) :- 
  vgenid(_), % Non-backtrackable predicate
  xrule(Schema/Id1:Cs1==>Ds1),
  xrule(Schema/Id2:Cs2==>Ds2),
  Id1 \= Id2, 
  subsumed_cond(Cs1,Cs2),    
  subsumed_dec(Ds1, Ds2),  
  vconcat(['\'In ', Schema,' rule: ',Id2,' subsumes rule ',Id1,'\''],RString), 
  vreport(0,[Id1,RString]),
  vreport(0,[Id2,RString]),
  fail.
vsubsume(_) :-  
  retractall(v(_)).

subsumed_cond([],[]) :- !. 
subsumed_cond([C1|Cs1],[C2|Cs2]) :-
  vcondition(C1, VC1),
  vcondition(C2, VC2),  
  sub(VC1,VC2),
  subsumed_cond(Cs1,Cs2).  
  
sub(C1,C2) :-
  retractall(v(_)),
  assert(v(C2)), 
  valid(C1).

subsumed_dec(Ds1, Ds2) :- 
  split_decision(Ds1,[D1,_,_,_]), 
  split_decision(Ds2,[D2,_,_,_]), 
  subsumed_dec_attributes(D1,D2).

subsumed_dec_attributes([],[]) :- !.
subsumed_dec_attributes([D1|Ds1],[D2|Ds2]) :- 
  sub(D1,D2),
  subsumed_dec_attributes(Ds1,Ds2).  

%------------------------------------------------------------------------------%
%--- REDUCTION ----------------------------------------------------------------%
%------------------------------------------------------------------------------%
vreduce(Schema):-
  vgenid(_), % Non-backtrackable predicate
  xrule(Schema/Id1:Cs1==>Ds1),
  xrule(Schema/Id2:Cs2==>Ds2),
  Id1 \= Id2, 
  split_decision(Ds1,[D1,_,_,_]), 
  split_decision(Ds2,[D2,_,_,_]), 
  D1 == D2,
  join_conds(Cs1,Cs2,Cond),
  neqset(Cond,[]),
  vconcat(['\'In ', Schema,' rule: ',Id1,' can be joined with rule ',Id2,'\''],RString), 
  vreport(0,[Id1,RString]),
  vreport(0,[Id2,RString]),
  fail.
vreduce(_) :- 
  retractall(j(_)).

join_conds([],[],[]) :- !.
join_conds([C1|Cs1],[C2|Cs2],[C|Cond]) :- 
  vcondition(C1, VC1),
  vcondition(C2, VC2), 
  join_cond(VC1, VC2, C),  
  join_conds(Cs1,Cs2,Cond).
  
join_cond(C1,C2,C) :-
  retractall(j(_)),
  assert(j(C2)),
  join(C1, C).
  
  
